/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*  Analyse les sorties de 'lance_traitement'  */
/*  FRED 0200  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>

#define TailleLigne	1000

typedef struct
	{
	int num,t_train,t_test,in;
	float pp_noeud,pp_ref,gain;
	} type_corpus;

#define MaxCorpus	1000

type_corpus TablCorpus[MaxCorpus];

int NbCorpus;

void LisData(FILE *file)
{
char ch[TailleLigne],patt[TailleLigne],*pt;
int taille,num,i;
float pp_noeud,pp_ref,gain;

NbCorpus=0;

while(fgets(ch,TailleLigne,file))
 {
 if (strstr(ch,"train"))
  {
  if (sscanf(ch,"Corpus %d : taille(train)=%d",&num,&taille)!=2) { fprintf(stderr,"ERREUR dans : %s\n",ch); exit(0); }
  TablCorpus[NbCorpus].num=num;
  TablCorpus[NbCorpus].t_train=taille;
  TablCorpus[NbCorpus].in=0;
  NbCorpus++;
  if (NbCorpus==MaxCorpus)
   { fprintf(stderr,"ERREUR : MaxCorpus trop petite ....\n"); exit(0); }
  }
 else
  if (strstr(ch,"test"))
   {
   if (sscanf(ch,"Corpus %d : taille(test)=%d",&num,&taille)!=2) { fprintf(stderr,"ERREUR dans : %s\n",ch); exit(0); }
   fgets(ch,TailleLigne,file);
   sscanf(ch,"\tPerplexite : modele noeud = %f  -  modele reference = %f  -  gain = %f",
	&pp_noeud,&pp_ref,&gain);
   for(i=0;(i<NbCorpus)&&(TablCorpus[i].num!=num);i++);
   if (i==NbCorpus) { fprintf(stderr,"ERREUR : num=%d inconnu !!\n",num); exit(0); }
   TablCorpus[i].t_test=taille;
   TablCorpus[i].in=1;
   TablCorpus[i].pp_noeud=pp_noeud;
   TablCorpus[i].pp_ref=pp_ref;
   TablCorpus[i].gain=gain;
   }
 }
}

void AfficheData()
{
int i,j;
for(i=j=0;i<NbCorpus;i++)
 if (TablCorpus[i].in)
  printf("Corpus:%2d (n%2d) - train=%5d - test=%5d  - pp noeud=%5.2f - pp ref=%5.2f - gain=%5.2f\n",
   ++j,TablCorpus[i].num,TablCorpus[i].t_train,TablCorpus[i].t_test,
   TablCorpus[i].pp_noeud,TablCorpus[i].pp_ref,TablCorpus[i].gain);
}

void GainMoyen()
{
double somm_ref,somm_noeud,somm_gain,moy_ref,moy_noeud,p_ref,p_noeud;
int i,nb,somm_t_test,somm_t_train,t_poubelle_train,t_poubelle_test;
moy_ref=moy_noeud=somm_ref=somm_noeud=somm_gain=(double)0;
somm_t_test=somm_t_train=0;
for(i=nb=0;i<NbCorpus;i++)
 if (TablCorpus[i].in)
  {
  nb++;
  /*
  printf("\tpp_ref=%lf pp_noeud=%lf log(pp_ref)=%lf log(pp_noeud)=%lf\n",
  TablCorpus[i].pp_ref,TablCorpus[i].pp_noeud,log((double)TablCorpus[i].pp_ref),log((double)TablCorpus[i].pp_noeud)); 
  */

  somm_ref+=((double)TablCorpus[i].pp_ref);
  somm_noeud+=((double)TablCorpus[i].pp_noeud);
  moy_ref+=log((double)TablCorpus[i].pp_ref);
  moy_noeud+=log((double)TablCorpus[i].pp_noeud);
  somm_gain+=(double)TablCorpus[i].gain;
  somm_t_test+=TablCorpus[i].t_test;
  somm_t_train+=TablCorpus[i].t_train;
  if (TablCorpus[i].in==2)
   {
   t_poubelle_train=TablCorpus[i].t_train;
   t_poubelle_test=TablCorpus[i].t_test;
   }
  }
/*
printf("somm_ref=%lf somm_noeud=%lf\n",somm_ref,somm_noeud);
printf("moy_ref=%lf moy_noeud=%lf\n",moy_ref,moy_noeud);
*/
moy_ref=moy_ref/(double)nb;
moy_noeud=moy_noeud/(double)nb;
/*
printf("moy_ref=%lf moy_noeud=%lf\n",moy_ref,moy_noeud);
*/
p_ref=exp(moy_ref);
p_noeud=exp(moy_noeud);
/*
printf("p_noeud=%lf p_ref=%lf\n",p_noeud,p_ref);
*/

/*
printf("gain moyen = %.2lf - ref moyen = %.2lf - noeud moyen = %.2lf - diff = %.2lf\n",
 somm_gain/(double)nb,somm_ref/(double)nb,somm_noeud/(double)nb,(somm_ref/(double)nb)-(somm_noeud/(double)nb));
*/

printf("Taille moyenne des fichiers d'apprentissage (en nb de mots) = %d\n",somm_t_train/nb);
printf("Taille moyenne des fichiers de test (en nb de mots) = %d\n",somm_t_test/nb);
/*
printf("Pour le corpus 'poubelle' : train = %d - test = %d - train moy sans poub = %d - test moy sans poub = %d\n",
 t_poubelle_train,t_poubelle_test,(somm_t_train-t_poubelle_train)/(nb-1),(somm_t_test-t_poubelle_test)/(nb-1));
*/
printf("pp ref = %.2lf - pp noeud = %.2lf - gain = %.2lf (%.2lf%%)\n",
 p_ref,p_noeud,p_ref-p_noeud,((p_ref-p_noeud)*(double)100)/p_ref);

}

int main()
{
LisData(stdin);
AfficheData();

GainMoyen();
}
  
