/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  lia_sct : fonction et type generaux  */

#include <lia_sct.h>
#include <lia_sct_pplex.h>

/* Variables Globales */

int Debug=1;                             /* niveau de Debug */
int MaxHauteur;                          /* hauteur maxi de l'arbre construit ou lu */
char TablTypeQuestion[NbTypeQuestion]    /* code des questions : J L R U and G for global */
	= {'J','L','R','U','G'} ;
char TablDirection[NbDirections]         /* code des directions : r l */
	= {'r','l'} ;
double MaxCritere;			 			 /* quantite maximale pouvant recevoir le critere choisi */
int NUMBER_NODE=0;						 /* Node ID */

/*................................................................*/

/*  Fonctions diverses et variees  */

void ERREUR(ch1,ch2)
 char *ch1,*ch2;
{
fprintf(stderr,"ERREUR : %s %s\n",ch1,ch2);
exit(0);
}

void *my_malloc(t)
 int t;
{
void *pt;
pt=malloc(t);
if (pt==NULL) { fprintf(stderr,"sorry, no memory available .... (taille demandee=%d)\n",t); exit(0); }
return pt;
}

/*................................................................*/

/*  Allocation memoire  */

ty_node nouveau_node(char KS,char t_noeud,int cate,char t_question,
		     char qloc,wrd_index_t q,char niveau_q,double *proba)
{
ty_node pt;
pt=(ty_node)my_malloc(sizeof(struct type_node));
pt->KS=KS;
pt->t_noeud=t_noeud;
pt->cate=cate;
pt->t_question=t_question;
pt->qloc=qloc;
pt->q=q;
pt->niveau_q=niveau_q;
pt->fd=pt->fg=NULL;
pt->proba=proba;
return pt;
}

ty_item nouveau_item(wrd_index_t *code,char inKS,ty_item pred, ty_format my_format)
{
ty_item pt;
int i;
pt=(ty_item)my_malloc(sizeof(struct type_item));
if (code)
 {
 pt->code=(wrd_index_t *)my_malloc(sizeof(wrd_index_t)*my_format->nb_niveaux);
 for(i=0;i<my_format->nb_niveaux;i++) pt->code[i]=code[i];
 }
else pt->code=NULL;
pt->inKS=inKS;
pt->suiv=NULL;
pt->pred=pred;
return pt;
}

void libere_item(ty_item pt)
{
if (pt)
 {
 if (pt->code) free(pt->code);
 free(pt);
 }
}

void libere_list_item(ty_item pt)
{
ty_item ptbak;

while(pt)
 {
 ptbak=pt;
 pt=pt->suiv;
 libere_item(ptbak);
 }
}

ty_sample nouveau_sample(
 wrd_index_t ref,
 int numero,
 type_KS KS,
 ty_item item,
 double nb_occu,
 wrd_index_t *classe,
 wrd_index_t *questall,
 ty_sample suiv)
{
ty_sample pt;
pt=(ty_sample)my_malloc(sizeof(struct type_sample));
pt->ref=ref;
pt->numero=numero;
pt->KS=KS;
pt->item=item;
pt->nb_occu=nb_occu;
pt->classe=classe;
pt->questall=questall;
pt->suiv=suiv;
return pt;
}

ty_format nouveau_format(int nb_niveaux,
			 int nb_questall,
			 int indice_ML,
			 ty_lexique *tablex,
			 ty_lexique lex_reference,
			 ty_lexique lex_etiquette,
			 char mode,
			 char critere,
			 int nb_etiquette,
			 double delta)
{
ty_format pt;
pt=(ty_format)my_malloc(sizeof(struct type_format));
pt->nb_niveaux=nb_niveaux;
pt->nb_questall=nb_questall;
pt->indice_ML=indice_ML;
pt->tablex=tablex;
pt->lex_reference=lex_reference;
pt->lex_etiquette=lex_etiquette;
pt->mode=mode;
pt->critere=critere;
pt->nb_etiquette=nb_etiquette;
pt->delta=delta;
return pt;
}

void libere_format(ty_format pt)
{
if (pt->tablex) free(pt->tablex);
free(pt);
}

/* renvoi une chaine de caractere corrspondant a un code et un niveau de representation */
char *renvoi_chaine(wrd_index_t code, int niveau, ty_format my_format)
{
char *ch_graf;
if (!(Code2Mot(code,&ch_graf,my_format->tablex[niveau])))
 {
 fprintf(stderr,"C'est quoi ce code : %ld pour le niveau %d\n",code,niveau);
 exit(0);
 }
return ch_graf;
}

/*................................................................*/

/* fonction de gestion des exemples et des lexiques */

/* renvoi une chaine de caractere corrspondant a un code et un lexique */
char *renvoi_chaine_lex(wrd_index_t code,ty_lexique lexique)
{
char *ch_graf;
if (!(Code2Mot(code,&ch_graf,lexique)))
 {
 fprintf(stderr,"C'est quoi ce code (lex) : %ld\n",code);
 exit(0);
 }
return ch_graf;
}

/* code un item a partir d'une chaine de caractere representant un exemple */

ty_item CodeItem(char *ch2, ty_format my_format)
{
ty_item tete,pt;
int i;
wrd_index_t *code;
static char chaine[10*TailleLigne],*ch;

strcpy(chaine,ch2);

code=(wrd_index_t *)my_malloc(sizeof(wrd_index_t)*my_format->nb_niveaux);
tete=pt=nouveau_item(NULL,0,NULL,my_format);
ch=strtok(chaine," \t\n");
while (ch)
 {
 for(i=0;i<my_format->nb_niveaux;i++)
  {
  if (!Mot2Code(ch,&(code[i]),my_format->tablex[i]))
   {
   /* ERREUR("Mauvais code pour l'item :",ch); ON MET <UNK> A LA PLACE */
   /*fprintf(stderr,"Warning : le mot '%s' est inconnu du lexique niveau %d\n",ch,i);*/

   if (!Mot2Code("<UNK>",&(code[i]),my_format->tablex[i])) ERREUR("Pas de <UNK> dans le lexique","");
   }
  if ((ch=strtok(NULL," \t\n"))==NULL)
   if (i<(my_format->nb_niveaux-1))  ERREUR("Mauvais format ds le fichier corpus :",ch2);
  }
 pt->suiv=nouveau_item(code,0,pt,my_format);
 pt=pt->suiv;
 }
pt=tete; tete=tete->suiv; tete->pred=NULL; free(pt);
free(code);
return tete;
}

/*................................................................*/

/*  Fonctions d'affichage  */

void AfficheArbre(ty_node pt, FILE *file, ty_format my_format)
{
if (pt)
 {
 if (pt->q)
  fprintf(file,"%c %d %c %s %d %c %c\n",pt->t_noeud,pt->cate,pt->KS,
	renvoi_chaine(pt->q,pt->t_question=='G'?my_format->nb_niveaux:pt->niveau_q,my_format),
	(int)pt->niveau_q+1,pt->t_question,pt->qloc);
 else
  fprintf(file,"%c %d %c NULL 1 %c %c\n",pt->t_noeud,pt->cate,pt->KS,pt->t_question,pt->qloc);
 if (pt->proba)
  {
  int i;
  for(i=0;i<my_format->nb_etiquette;i++)
   {
   fprintf(file,"%s %.4lf",renvoi_chaine_lex(i+1,my_format->lex_etiquette),pt->proba[i]);
   if (i<my_format->nb_etiquette-1) fprintf(file," "); else fprintf(file,"\n");
   }
  }
 AfficheArbre(pt->fg,file,my_format);
 AfficheArbre(pt->fd,file,my_format);
 }
}

void AfficheListSample(ty_sample pt, ty_format my_format)
{
int n;
ty_item poin;

for(n=0;pt;pt=pt->suiv,n++)
 {
 if (pt->KS.ptKS==NULL) fprintf(stdout,"%d] <1,NULL,1> ",n+1);
 else fprintf(stdout,"%d] <%d,%s,%d> ",n+1,pt->KS.gap_avant,
	renvoi_chaine(pt->KS.ptKS->code[0],0,my_format),pt->KS.gap_apres);
 for(poin=pt->KS.ptKS;poin;poin=poin->suiv) fprintf(stdout," %s",renvoi_chaine(poin->code[0],0,my_format));
 printf("\n");
 }
}

void affiche_item(ty_item pt, ty_format my_format)
{
int i,j=0;
char *ch;

while(pt)
 {
 printf(" [");
 for(i=0;i<my_format->nb_niveaux;i++)
  {
  ch=renvoi_chaine(pt->code[i],i,my_format);
  if (i<my_format->nb_niveaux-1) printf("%s,",ch); else  printf("%s]",ch);
  j+=strlen(ch);
  }
 pt=pt->suiv;
 if (j>60) { printf("\n\t"); j=0; }
 }
printf("\n");
}

/*................................................................*/

/* lecture du fichier de description des formats */

ty_format LectureDescripteurFormat(char *nomfic)
{
FILE *file,*file_eti,*file_test;
char ch[TailleLigne],*pt,chtest[TailleLigne];
signed char mode=-1,critere=-1;
ty_lexique *tablex=NULL,lex_reference=NULL,lex_etiquette=NULL;
int nb_niveaux=0,nb_questall=0,indice_ML=0,i,nb_etiquette=0,si_delta=0;
double delta;

if (!(file=fopen(nomfic,"rt"))) ERREUR("can't open : ",nomfic);

while (fgets(ch,TailleLigne,file))
 {
 if (!strncmp(ch,"NB_NIVEAUX",10))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%d",&nb_niveaux)!=1) ERREUR("Mauvais format1 dans le descripteur","");
  } else
 if (!strncmp(ch,"NB_MAX_GLB_QUEST",16))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%d",&nb_questall)!=1) ERREUR("Mauvais format_XX1 dans le descripteur","");
  } else
 if (!strncmp(ch,"INDICE_ML",9))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%d",&indice_ML)!=1) ERREUR("Mauvais format2 dans le descripteur","");
  } else
 if (!strncmp(ch,"LEXIQUE",7))
  {
  if (tablex==NULL)
   {
   tablex=(ty_lexique *)my_malloc(sizeof(ty_lexique)*(nb_niveaux+1)); /* the global quest are stored in nb_niveaux+1 */
   for(i=0;i<nb_niveaux+1;i++) tablex[i]=NULL;
   }
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (!strcmp(pt,"REF"))
   {
   pt=strtok(NULL," \t\n");
   /* now the lex reference is optional => big and rarely useful */
   if (file_test=fopen(pt,"rt"))
    {
    fclose(file_test);
    lex_reference=ChargeLexique(pt);
    }
   }
  else
   if (!strcmp(pt,"ETI"))
    {
    pt=strtok(NULL," \t\n");
    lex_etiquette=ChargeLexique(pt);
    if (!(file_eti=fopen(pt,"rt"))) ERREUR("can't open :",pt);
    for(i=0;fgets(chtest,TailleLigne,file_eti);i++);
    fclose(file_eti);
    nb_etiquette=i;
    }
   else
    {
    if (sscanf(pt,"%d",&i)!=1) ERREUR("Mauvais format3 dans le descripteur","");
    pt=strtok(NULL," \t\n");
    if (tablex) tablex[i-1]=ChargeLexique(pt);
    else ERREUR("Dans le descripteur, le nb de niveaux doit etre defini en premier","");
    }
  } else
 if (!strncmp(ch,"GLOBAL_QUEST",12))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (tablex) tablex[nb_niveaux]=ChargeLexique(pt); /* stored in tablex[nb_niveaux] so nb_niveaux+1 position */
  else ERREUR("Dans le descripteur, le nb de niveaux doit etre defini en premier","");
  }
 else
 if (!strncmp(ch,"MODE",4))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (!strcmp(pt,"SUPERVISE"))
   {
   mode=SUPERVISE;
   }
  else
   if (!strcmp(pt,"NON_SUPERVISE"))
    {
    mode=NON_SUPERVISE;
    }
   else ERREUR("mode inconnu :",pt);
  } else
 if (!strncmp(ch,"CRITERE",7))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (!strcmp(pt,"PPLEX"))
   {
   critere=PPLEX;
   MaxCritere=MAX_PPLEX;
   }
  else
  if (!strcmp(pt,"GINI"))
   {
   critere=GINI;
   MaxCritere=(double)1000; /*MAX_GINI*/
   } else
  if (!strcmp(pt,"GINI_LIGHT"))
   {
   critere=GINI_LIGHT;
   MaxCritere=(double)1000; /*MAX_GINI*/
   }
  else ERREUR("critere inconnu :",pt);
  } else
 if (!strncmp(ch,"DELTA",5))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%lf",&delta)!=1) ERREUR("Mauvais DELTA :",pt);
  si_delta=1;
  }
 }

if (mode==-1) ERREUR("Il manque le mode d'apprentissage de l'arbre","");
if (critere==-1) ERREUR("Il manque le critere de selection","");
if (tablex==NULL) ERREUR("Il manque le nb de niveaux dans le descripteur","");
if (indice_ML==0) ERREUR("Il manque l'indice du ML dans le descripteur","");

for(i=0;i<nb_niveaux;i++)
 if (tablex[i]==NULL) ERREUR("Descripteur -> il manque un fichier lexique","");
/* now the lex reference is optional => big and rarely useful */
/*if (lex_reference==NULL) ERREUR("Descripteur -> il manque le lexique reference","");*/
if (si_delta==0) ERREUR("Descripteur -> il manque le DELTA","");

return nouveau_format(nb_niveaux,nb_questall,indice_ML-1,tablex,lex_reference,lex_etiquette,mode,critere,nb_etiquette,delta);

}
 
