/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*  Permet de determiner un ou plusieurs corpus a 
    partir d'un exemple - Premiere implementation de TreeML
    Les exemples sont lus dans un fichier nbest (au format fmt)
    ou pour chaque idenficateur de phrase :
    - le premier representant est la reference
    - le deuxieme est l'hypothese du CNET
    - le 3eme est le 1best
    - le 4eme est le 2best
    etc  */
/*  FRED 0301  */

#include <lia_sct.h>

/*................................................................*/

#define CHOIX_ARBRE	1
#define CHOIX_BEST	2
#define CHOIX_CNET	3
#define CHOIX_REFE	4

int ChoixMethode=CHOIX_BEST;

/* stockage des items reference (pour le calcul du gain) et nbest */

#define TAILLE_NBEST	32

#define I_REFE	0
#define I_CNET	1
#define I_BEST	2

typedef struct type_itemref
	{
	char *ident;
	int ordre; /* stocke l'ordre de lecture, pour les resortir dans l'ordre .... (pour Yannick) */
	ty_item item[I_BEST+TAILLE_NBEST]; /* 0=REF 1=CNET 2=1BEST 3=2BEST .... */
	struct type_itemref *fg,*fd;
	} *ty_itemref;

ty_itemref ArbreNbest;
ty_itemref *TablNbest;

int NbItemref=0,NbNbest=0;

ty_itemref nouveau_itemref(char *ident,ty_item item)
{
ty_itemref pt;
int i;
pt=(ty_itemref)malloc(sizeof(struct type_itemref));
pt->ident=(char*)strdup(ident);
pt->item[0]=item;
for(i=1;i<I_BEST+TAILLE_NBEST;i++) pt->item[i]=NULL;
pt->fg=pt->fd=NULL;
pt->ordre=NbItemref;
NbItemref++;
return pt;
}

ty_itemref ajoute_itemref(ty_itemref racine,char *ident,ty_item item)
{
if (racine==NULL) racine=nouveau_itemref(ident,item);
else
 {
 int comp;
 comp=strcmp(ident,racine->ident);
 if (comp<0) racine->fg=ajoute_itemref(racine->fg,ident,item); else
 if (comp>0) racine->fd=ajoute_itemref(racine->fd,ident,item);
 else
  {
  /* on l'ajoute a la liste de nbest */
  int i;
  for(i=0;(i<I_BEST+TAILLE_NBEST)&&(racine->item[i]!=NULL);i++);
  if (i==I_BEST+TAILLE_NBEST) ERREUR("TAILLE_NBEST trop petite ??","");
  racine->item[i]=item;
  }
 }
return racine;
}

ty_item *recherche_itemref(ty_itemref racine,char *ident)
{
if (racine==NULL) return NULL;
else
 {
 int comp;
 comp=strcmp(ident,racine->ident);
 if (comp<0) return recherche_itemref(racine->fg,ident); else
 if (comp>0) return recherche_itemref(racine->fd,ident);
 else return racine->item;
 }
}

ty_itemref charge_itemref(char *nomfich,ty_format my_format)
{
FILE *file;
int n,i;
char ch[10*TailleLigne],*graf,*chitem;
ty_itemref racine=NULL;
ty_item item;

if (!(file=fopen(nomfich,"rt"))) ERREUR("can't open ",nomfich);
for(n=0;fgets(ch,10*TailleLigne,file);n++)
 if (ch[0]!='@')
  {
  if ((n+1)%5000==0) fprintf(stderr,"En cours : %d\n",n+1);
  graf=(char*)strtok(ch," \t\n");
  for(i=strlen(graf)+1;(graf[i]==' ')||(graf[i]=='\t');i++); chitem=graf+i;
  item=CodeItem(chitem,my_format);
  racine=ajoute_itemref(racine,graf,item);
  }
fclose(file);
return racine;
}

void range_itemref_tabl(ty_itemref racine,ty_itemref *tabl)
{
if (racine)
 {
 if (racine->ordre>=NbNbest) ERREUR("pb dans l'ordre des item","");
 tabl[racine->ordre]=racine;
 range_itemref_tabl(racine->fg,tabl);
 range_itemref_tabl(racine->fd,tabl);
 }
}

/*................................................................*/

/*  Codage de la repartition des exemples apres le parcours de l'arbre  */

#define MAX_PPLEX	(double)10000000

typedef struct
	{
	int i_ml; /* contient les indices des ML ou bien -1 */
	ty_node node; /* adresse du noeud */
	} type_tree_ML;

type_tree_ML *TreeML; /* contient les indices des ML ou bien -1 */
int NbML;
int I_ModeleGeneral;	/* indice du modele general */

/*................................................................*/

int trouve_ML(ty_node pt)
{
int i;
for(i=0;(i<NbML)&&(TreeML[i].node!=pt);i++);
if (i==NbML) return -1;
if (TreeML[i].i_ml==-1) ERREUR("ML non charge ??","");
return TreeML[i].i_ml;
}

int trouve_indice_noeud(int i_ml)
{
int i;
for(i=0;(i<NbML)&&(TreeML[i].i_ml!=i_ml);i++);
if (i==NbML) ERREUR("pas normal","");
return i;
}

/*................................................................*/

/*  Parcours du CART avec un item  */

#define MAX_ITEM		200

void SauvEtat(pt,etat)
 ty_item pt;
 char *etat;
{
int n;
for(n=0;(n<MAX_ITEM)&&(pt);pt=pt->suiv,n++)
 etat[n]=pt->inKS;
if (n==MAX_ITEM) ERREUR("Cste MAX_ITEM trop petite !!!!","");
}

void SauvEtatNBEST(ty_item *item, char etat[TAILLE_NBEST][MAX_ITEM])
{
int n,i;
ty_item pt;

for(i=0;i<TAILLE_NBEST;i++)
 {
 for(n=0,pt=item[i];(n<MAX_ITEM)&&(pt);pt=pt->suiv,n++) etat[i][n]=pt->inKS;
 if (n==MAX_ITEM) ERREUR("Cste MAX_ITEM trop petite !!!!","");
 }
}

void RestorEtat(pt,etat)
 ty_item pt;
 char *etat;
{
int n;
for(n=0;(n<MAX_ITEM)&&(pt);pt=pt->suiv,n++)
 pt->inKS=etat[n];
if (n==MAX_ITEM) ERREUR("Cste MAX_ITEM trop petite !!!!","");
}

void RestorEtatNBEST(ty_item *item, char etat[TAILLE_NBEST][MAX_ITEM])
{
ty_item pt;
int n,i;

for(i=0;i<TAILLE_NBEST;i++)
 {
 for(n=0,pt=item[i];(n<MAX_ITEM)&&(pt);pt=pt->suiv,n++) pt->inKS=etat[i][n];
 if (n==MAX_ITEM) ERREUR("Cste MAX_ITEM trop petite !!!!","");
 }
}

ty_item Compatible(ty_item item,type_KS KS,wrd_index_t question,char niveau_q,char qloc,char gap_avant,char gap_apres)
{
ty_item pt;

if (qloc=='l')  /* on cherche a la gauche du KS */
 {   /* sur la racine, ptKS est egal a NULL */
 /* test de coherence */
 /* s'il n'y a pas de GAP avant, on recule vers la gauche jusqu'a en trouver un */
 if (KS.gap_avant==0)
  {
  for(pt=KS.ptKS;(pt)&&(pt->inKS==1);pt=pt->pred);
  if (pt==NULL) return NULL; /*ERREUR("Il n'y a pas de GAP sur la gauche !!!!","");*/
  }
 else if (KS.ptKS==NULL) pt=item; else pt=KS.ptKS->pred;

 if (pt==NULL) return NULL;

 if (gap_apres==0)
  { /* la question doit se trouver juste avant le KS */
  if ((pt->inKS==0)&&(question==pt->code[niveau_q]))
   { pt->inKS=1; return pt; }
  else return NULL;
  }
 else
  { /* il peut y avoir un gap entre la question et le KS */
  /* MODIF : il DOIT y avoir un gap */
  pt=pt->pred;
  while ((pt)&&(pt->inKS==0)&&(question!=pt->code[niveau_q])) pt=pt->pred;
  if ((pt)&&(pt->inKS==0))
   { pt->inKS=1; return pt; }
  else return NULL;
  }
 }
else
 {  /* on cherche a la droite du KS */
 /* test de coherence */
 /* s'il n'y a pas de GAP apres, on avance vers la droite jusqu'a en trouver un */
 if (KS.gap_apres==0)
  {
  for(pt=KS.ptKS==NULL?item:KS.ptKS;(pt)&&(pt->inKS==1);pt=pt->suiv);
  if (pt==NULL) return NULL; /*ERREUR("Il n'y a pas de GAP sur la droite !!!!","");*/
  }
 else if (KS.ptKS==NULL) pt=item; else pt=KS.ptKS->suiv;

 if (pt==NULL) return NULL;

 if (gap_avant==0)
  { /* la question doit se trouver juste apres le KS */
  if ((pt->inKS==0)&&(question==pt->code[niveau_q]))
   { pt->inKS=1; return pt; }
  else return NULL;
  }
 else
  { /* il peut y avoir un gap entre la question et le KS */
  /* MODIF : il DOIT y avoir un gap */
  pt=pt->suiv;
  while ((pt)&&(pt->inKS==0)&&(question!=pt->code[niveau_q])) pt=pt->suiv;
  if ((pt)&&(pt->inKS==0))
   { pt->inKS=1; return pt; }
  else return NULL;
  }
 }
}

ty_node Parcours(ty_node pt,ty_item item,type_KS KS,int *nb_yes,ty_format my_format)
{
char etat[MAX_ITEM];
ty_item resu;
type_KS newKS;
static char Debug_quest[TailleLigne];

if (pt)
 {
 if (pt->t_noeud=='L')
  {
  if (Debug>=2) fprintf(stdout,"Categorie : %c\n",pt->cate);
  return pt;
  }
 else
  {
  /*  S'il n'y a pas de GAP, on s'arrete  */
  /*
  if ((KS.gap_avant==0)&&(KS.gap_apres==0))
   {
   fprintf(stdout,"Warning : pas de GAP -> queskon renvoit ????\n");
   return;
   }
  */

  SauvEtat(item,etat);

  if (Debug>=2) { fprintf(stdout,"item = "); affiche_item(item,my_format); }
  switch (pt->t_question)
   {
   case 'J' :  /* < ? > */
	if (Debug>=2) sprintf(Debug_quest,"<%s>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,0,0))
	 { newKS.gap_avant=newKS.gap_apres=0; newKS.ptKS=resu; }
	break;
   case 'L' : /* < ?+> */
	if (Debug>=2) sprintf(Debug_quest,"<%s+>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,0,1))
	 { newKS.gap_avant=0; newKS.gap_apres=1; newKS.ptKS=resu; }
	break;
   case 'R' : /* <+? > */
	if (Debug>=2) sprintf(Debug_quest,"<+%s>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,1,0))
	 { newKS.gap_avant=1; newKS.gap_apres=0; newKS.ptKS=resu; }
	break;
   case 'U' : /* <+?+> */
	if (Debug>=2) sprintf(Debug_quest,"<+%s+>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,1,1))
	 { newKS.gap_avant=newKS.gap_apres=1; newKS.ptKS=resu; }
	break;
   default  :
	fprintf(stdout,"Je ne connais pas la question : %c\n",pt->t_question);
	exit(0);
   }
  if (resu)
   {
   if (Debug>=2)
    {
    fprintf(stdout,"  Oui a la question : %s sur le GAP %c\n",Debug_quest,pt->qloc);
    fprintf(stdout,"    avec KS=<%c,%s,%c>\n",'0'+KS.gap_avant,
	KS.ptKS!=NULL?renvoi_chaine((KS.ptKS)->code[my_format->indice_ML],my_format->indice_ML,my_format):"NULL",'0'+KS.gap_apres);
    }
   (*nb_yes)++;
   return Parcours(pt->fg,item,newKS,nb_yes,my_format);
   }
  else
   {
   if (Debug>=2)
    {
    fprintf(stdout,"  Non a la question : %s sur le GAP %c\n",Debug_quest,pt->qloc);
    fprintf(stdout,"    avec KS=<%c,%s,%c>\n",'0'+KS.gap_avant,
	KS.ptKS!=NULL?renvoi_chaine((KS.ptKS)->code[my_format->indice_ML],my_format->indice_ML,my_format):"NULL",'0'+KS.gap_apres);
    }
   RestorEtat(item,etat);
   return Parcours(pt->fd,item,KS,nb_yes,my_format);
   }
  }
 }
return NULL;
}

/*................................................................*/

/*  Lecture d'un fichier CART et construction de l'arbre  */

ty_node ConstruitArbre(FILE *file, char KS, ty_format my_format)
{
char ch[TailleLigne],t_noeud,t_question,qloc,q[TailleLigne],old_KS,niveau_q,*etiq,*pb;
ty_node pt;
wrd_index_t code_q,i_etiq;
int i,int_niveau_q,cate;
double *proba;

if (fgets(ch,TailleLigne,file))
 {
 i=sscanf(ch,"%c %d %c %s %d %c %c",&t_noeud,&cate,&old_KS,q,&int_niveau_q,&t_question,&qloc);
 if (i!=7)
  {
  sprintf(q,"%d elements trouves dans : %s (au lieu de 7)",i,ch);
  ERREUR("Mauvais format dans le fichier arbre",q);
  }

 niveau_q=(char)int_niveau_q-1;

 if (!strcmp(q,"NULL")) code_q=0; /* on mets le <UNK> sur NULL */
 else
  if (!Mot2Code(q,&code_q,my_format->tablex[niveau_q])) ERREUR("Item inconnu : ",q);

 proba=NULL;

 if (t_noeud=='L') /* on est sur une feuille */
  {
  if (my_format->mode==SUPERVISE) /* il doit y avoir une repartition de probas */
   {
   if (!fgets(ch,TailleLigne,file)) ERREUR("Pas de repart de proba ds l'arbre","");
   proba=(double*)malloc(sizeof(double)*my_format->nb_etiquette);
   for(i=0;i<my_format->nb_etiquette;i++)
    {
    if (i==0) etiq=strtok(ch," \t\n"); else etiq=strtok(NULL," \t\n");
    if (etiq==NULL) ERREUR("repart de proba non valide","");
    pb=strtok(NULL," \t\n");
    if (pb==NULL) ERREUR("repart de proba non valide","");
    if (!Mot2Code(etiq,&i_etiq,my_format->lex_etiquette)) ERREUR(etiq,"inconnu");
    if (sscanf(pb,"%lf",&(proba[i_etiq-1]))!=1) ERREUR("repart de proba non valide","");
    }
   }
  }

 pt=nouveau_node(KS,t_noeud,cate,t_question,qloc,code_q,niveau_q,proba);

 NbML++; /* on rajoute un ML car on rajoute un noeud */

 if (t_noeud=='I')
  {
  pt->fg=ConstruitArbre(file,t_question,my_format);
  pt->fd=ConstruitArbre(file,KS,my_format);
  }
 return pt;
 }
return NULL;
}

ty_node LectureTree(char *chfile,ty_format my_format)
{
FILE *file;
int n;
ty_node racine;

if (!(file=fopen(chfile,"rt"))) ERREUR("Can't open : ",chfile);

/*  La KS a la racine est X cad <+>  */
NbML=0;
racine=ConstruitArbre(file,'X',my_format);
fclose(file);

/* on alloue le tableau des ML */
TreeML=(type_tree_ML*)malloc(sizeof(type_tree_ML)*NbML);
for(n=0;n<NbML;n++) { TreeML[n].i_ml=-1; TreeML[n].node=NULL; }

if (Debug>=2) printf("\tTreeML va contenir %d ML au maximum\n",NbML);
return racine;
}

/*................................................................*/

/*  Gere les noeud_MLs d'exemples dans les feuilles  */

void ecris_item_niveau(char *ref,char *etiquette,ty_item item,int niveau,FILE *file,ty_format my_format)
{
fprintf(file,"%s",ref);
while (item)
 {
 fprintf(file," %s",renvoi_chaine(item->code[niveau-1],niveau-1,my_format));
 item=item->suiv;
 }
fprintf(file," => %s\n",etiquette);
}

void ecris_item(char *ref,ty_item item,FILE *file,ty_format my_format)
{
int i;
fprintf(file,"%s",ref);
while (item)
 {
 for(i=0;i<my_format->nb_niveaux;i++) fprintf(file," %s",renvoi_chaine(item->code[i],i,my_format));
 item=item->suiv;
 }
fprintf(file,"\n");
}

/*................................................................*/

/* Lecture du corpus et construction des structures des ngrams */
/* MODIF : lecture de fichiers au format :

i=item (l items), n=niveau (k niveaux)

<graf> <i1 n1> <i1 n2> .... <i1 nk> <i2 n1> <i2 n2> .... <i2 nk> .... <ilnk>

*/

ty_node test_item(ty_item item, ty_node racine, ty_format my_format)
{
ty_node feuille;
type_KS KS;
int nb_yes;

if (Debug>=2)
 { printf("On traite :\n"); if (item) affiche_item(item,my_format); else printf("NULL ????\n"); }

/*  Initialisation du KS  */
KS.gap_avant=KS.gap_apres=1;
KS.ptKS=NULL;

/*  Parcours de l'arbre de decision  */
nb_yes=0;
feuille=Parcours(racine,item,KS,&nb_yes,my_format);

if (feuille==NULL) printf("L'item n'est pas analysee par le CART\n");

return feuille;
}

/*................................................................*/

/* calcule de la perplexite */

double calcule_pplex(ty_item item,int i_ML,ty_format my_format)
{
double total,pb,pplex;
int nb;
ty_item pt;
err_t retour;

for(total=(double)0,nb=0,pt=item;pt->suiv;pt=pt->suiv,nb++)
 {
 retour=gram_proba_to_bigram(&pb,pt->code[my_format->indice_ML],pt->suiv->code[my_format->indice_ML],i_ML,0);
 if (retour==GRAM_CODE_UNKNOWN)
  retour=gram_proba_to_bigram(&pb,0,pt->suiv->code[my_format->indice_ML],i_ML,0);
 if (retour==GRAM_CODE_UNKNOWN)
  retour=gram_proba_to_bigram(&pb,pt->code[my_format->indice_ML],0,i_ML,0);
 if (retour==GRAM_CODE_UNKNOWN)
  retour=gram_proba_to_bigram(&pb,0,0,i_ML,0);
 if (retour!=CORRECT) { fprintf(stderr,"ERREUR : retour gram_proba_to_bigram = %d\n",retour); exit(0); }
 total+=(-pb);
 }
pplex=exp(total/(double)nb*log(10.0));
return pplex;
}

#define INCLUDE_UNKS	True

double somm_proba(ty_item item,int i_ML,int *nbbig,int si_filtre,ty_format my_format)
{
double total,pb;
int nb;
ty_item pt;
err_t retour;
wrd_index_t m0,m1;
float lp,fr;

if (!strcmp(renvoi_chaine(item->code[0],0,my_format),"ETIQ")) { item=item->suiv; /* on saute le premier si c'est l'etiquette */ }

for(total=(double)0,nb=0,pt=item;pt->suiv;pt=pt->suiv)
 {
 m0=pt->code[my_format->indice_ML];
 m1=pt->suiv->code[my_format->indice_ML];
 /* si l'un des deux est un mot <UNK> (BIZARRE) on le compte quand meme si INCLUDE_UNKS est a vrai */
 if ((si_filtre==0)||(Recherche2Gram(m0,m1,&lp,&fr,ML(i_ML)))||((INCLUDE_UNKS)&&((m0==0)||(m1==0))))
  {
  /*if ((INCLUDE_UNKS)||((Recherche1Gram(m1,&lp,&fr,ML(i_ML)))&&(Recherche1Gram(m0,&lp,&fr,ML(i_ML)))))*/

  if ((INCLUDE_UNKS)||((m0!=0)&&(m1!=0)))
   {
   retour=gram_proba_to_bigram(&pb,m0,m1,i_ML,0);
   if (retour!=CORRECT) { fprintf(stderr,"ERREUR : retour gram_proba_to_bigram = %d\n",retour); exit(0); }
   total+=(-pb);
   nb++;
   }
  }
 else
  { total+=(8.0); nb++; } /* on penalise le backoff fortement */
 }
(*nbbig)=nb;
return total;
}

#define MAX_BIG	400

typedef struct
	{
	wrd_index_t m0,m1;
	int nb;
	} type_stockbig;

double somm_proba_nbest(ty_item *tablitem,int i_ML,int *nbbigutil,ty_format my_format)
{
double total,pb;
float lp,fr;
int i,j,nb,nbbig;
ty_item pt,item;
wrd_index_t m0,m1;
type_stockbig tablbig[MAX_BIG];
err_t retour;

for (i=I_CNET,nbbig=nb=0;i<I_BEST+2/*TAILLE_NBEST*/;i++)
 if (tablitem[i])
  {
  nb++;
  item=tablitem[i];
  if (!strcmp(renvoi_chaine(item->code[0],0,my_format),"ETIQ")) { item=item->suiv; /* on saute le premier si c'est l'etiquette */ }
  for(pt=item;pt->suiv;pt=pt->suiv)
   {
   m0=pt->code[my_format->indice_ML];
   m1=pt->suiv->code[my_format->indice_ML];
   for(j=0;(j<nbbig)&&((tablbig[j].m0!=m0)||(tablbig[j].m1!=m1));j++);
   if (j==nbbig)
    {
    tablbig[j].m0=m0; tablbig[j].m1=m1; tablbig[j].nb=TAILLE_NBEST/nb;
    nbbig++;
    if (nbbig==MAX_BIG) ERREUR("MAX_BIG trop petit","");
    } else tablbig[j].nb+=TAILLE_NBEST/nb;
   }
  }
for(i=0,total=(double)0,(*nbbigutil)=0;i<nbbig;i++)
 if (tablbig[i].nb>=8)
  if (Recherche2Gram(tablbig[i].m0,tablbig[i].m1,&lp,&fr,ML(i_ML)))
   {
   retour=gram_proba_to_bigram(&pb,tablbig[i].m0,tablbig[i].m1,i_ML,0);
   if (retour!=CORRECT) { fprintf(stderr,"ERREUR : retour gram_proba_to_bigram = %d\n",retour); exit(0); }
   total+=(-pb)*(double)tablbig[i].nb;
   (*nbbigutil)+=tablbig[i].nb;
   }
  else
   { total+=(8.0)*(double)tablbig[i].nb; (*nbbigutil)+=tablbig[i].nb; } /* on penalise le backoff fortement */

if (Debug>=2) printf("Taille nbest=%d nbbigutil=%d\n",nb,*nbbigutil);
return total;
}

double calcule_pplex_nbest(ty_item *item,int i_ML,double *pb_cnet,double *pb_refe,ty_format my_format)
{
double total,pplex;
int nb,nbbig,i;

/* on fait la somme sur le nbest */

if (0)
 total=somm_proba_nbest(item,i_ML,&nbbig,my_format);
else
 for(i=I_BEST,nbbig=0,total=(double)0;i<I_BEST+1/*TAILLE_NBEST*/;i++)
  if (item[i])
   {
   total+=somm_proba(item[i],i_ML,&nb,1/* on ne filtre pas pour les concepts */,my_format);
   nbbig+=nb;
   }

pplex=exp(total/(double)nbbig*log(10.0));

/* on fait la somme sur le CNET */
for(i=I_CNET,nbbig=0,total=(double)0;i<I_CNET+1;i++)
 if (item[i])
  {
  total+=somm_proba(item[i],i_ML,&nb,1,my_format);
  nbbig+=nb;
  }
(*pb_cnet)=exp(total/(double)nbbig*log(10.0));

/* on fait la somme sur la REFE */
for(i=I_REFE,nbbig=0,total=(double)0;i<I_REFE+1;i++)
 if (item[i])
  {
  total+=somm_proba(item[i],i_ML,&nb,1,my_format);
  nbbig+=nb;
  }
(*pb_refe)=exp(total/(double)nbbig*log(10.0));

return pplex;
}

/*................................................................*/

/* parcours l'arbre en fonction de la pplex et des KS */

int NbFeuilleAtteinte, /* nb de feuilles atteintes dans la recherche d'un sous-ML */
    OkDirection,       /* Booleen indiquant si les choix effectue par la reference
                          dans l'arbre et ceux du ML noeud sont identiques */
    NbChoixOui;	       /* nb de fois ou un exemple reference a repondu 'OUI' dans l'arbre */

ty_node choix_direction(ty_node pt,double pplex_noeud, double pplex_g,double pplex_d)
{
double diff_refe,diff_cnet,diff_best,min_best,min_cnet,min_refe,minimum,difference;
ty_node gain_pplex;

minimum=min(pplex_g,pplex_d); difference=pplex_g-pplex_d;

if (Debug>=2) printf("\tchoix_direction: minimum=%.2lf difference=%.2lf\n",minimum,difference);

if (minimum>pplex_noeud) gain_pplex=NULL;
else gain_pplex=difference<0?pt->fg:pt->fd;

return gain_pplex;
}

ty_item pose_question(ty_item item,type_KS KS,ty_node pt,type_KS *newKS,ty_format my_format)
{
ty_item resu=NULL;

if (Debug>=2) { printf("on pose la question sur l'item : "); affiche_item(item,my_format); }

switch (pt->t_question)
 {
 case 'J' :  /* < ? > */
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,0,0))
	 { newKS->gap_avant=newKS->gap_apres=0; newKS->ptKS=resu; }
	break;
 case 'L' : /* < ?+> */
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,0,1))
	 { newKS->gap_avant=0; newKS->gap_apres=1; newKS->ptKS=resu; }
	break;
 case 'R' : /* <+? > */
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,1,0))
	 { newKS->gap_avant=1; newKS->gap_apres=0; newKS->ptKS=resu; }
	break;
 case 'U' : /* <+?+> */
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,1,1))
	 { newKS->gap_avant=newKS->gap_apres=1; newKS->ptKS=resu; }
	break;
 default  :
	fprintf(stdout,"Je ne connais pas la question : %c\n",pt->t_question);
	exit(0);
 }
if (Debug>=2) printf("%s\n",resu==NULL?"BAAD":"GOOD");
return resu;
}

#define DECA		1

ty_node parcoursML(ty_node pt, ty_item *item, type_KS *tablKS,double pplex_noeud,
                   int niveau_max,int niveau,int indice_depart,ty_format my_format)
{
char etat[TAILLE_NBEST][MAX_ITEM];
ty_item resu,resu2;
ty_node choix_arbre,choix_pplex,gain_pplex,retour_noeud;
type_KS *newKS;
double pplex_g,pplex_d,pp_cnet_g,pp_cnet_d,pp_refe_g,pp_refe_d;
int i_ML_gauche,i_ML_droite,indice,nb_oui,nb_non,i_oui,i_non,i;

pplex_noeud=ceil(pplex_noeud)+0;

if (pt)
 {
 if (trouve_ML(pt)==-1) fprintf(stderr,"ATTENTION : -1  niveau=%d\n",niveau);

 if (pt->t_noeud=='L') { NbFeuilleAtteinte++; return pt; }
 else
 if (niveau==niveau_max) return pt; /* on a atteint le niveau Maximum */
 else
  {
  /* evaluation des perplexites sur les sous-arbres */
  if ((pt->fg)&&((i_ML_gauche=trouve_ML(pt->fg))!=-1))
   pplex_g=calcule_pplex_nbest(item,i_ML_gauche,&pp_cnet_g,&pp_refe_g,my_format);
  else pplex_g=pp_cnet_g=pp_refe_g=MAX_PPLEX;

  if ((pt->fd)&&((i_ML_droite=trouve_ML(pt->fd))!=-1))
   pplex_d=calcule_pplex_nbest(item,i_ML_droite,&pp_cnet_d,&pp_refe_d,my_format);
  else pplex_d=pp_cnet_d=pp_refe_d=MAX_PPLEX;

  /* evaluation de parcours de l'arbre */
  newKS=(type_KS *)malloc(sizeof(type_KS)*TAILLE_NBEST);
  /*indice=indice_depart;*/  indice=indice_depart=0;
  SauvEtatNBEST(item,etat);
  for(i=0;i<TAILLE_NBEST;i++) { newKS[i].gap_avant=newKS[i].gap_apres=0; newKS[i].ptKS=NULL; }
  resu2=resu=pose_question(item[indice],tablKS[indice],pt,&(newKS[indice]),my_format);
  for (indice=indice+1;(indice<indice_depart+DECA)&&(((resu2)&&(resu))||((!resu)&&(!resu2)));indice++)
   if (item[indice]) resu2=pose_question(item[indice],tablKS[indice],pt,&(newKS[indice]),my_format);
  if (((resu)&&(!resu2))||((!resu)&&(resu2))) /* si il y a divergence, alors on ne descend plus */
   {
   /* alors on s'arrete */
   resu=NULL; choix_arbre=NULL;
   }
  else /* on choisit la direction */
   {
   if ((resu)&&(pplex_g!=MAX_PPLEX)) choix_arbre=pt->fg;
   else
    if ((resu==NULL)&&(pplex_d!=MAX_PPLEX)) choix_arbre=pt->fd;
    else choix_arbre=NULL;
   }

  if (pt->fg==NULL) ERREUR("Pourquoi fg==NULL??","");
  if (pt->fd==NULL) ERREUR("Pourquoi fd==NULL??","");

  if (ChoixMethode==CHOIX_CNET) { pplex_g=pp_cnet_g; pplex_d=pp_cnet_d; } else
  if (ChoixMethode==CHOIX_REFE) { pplex_g=pp_refe_g; pplex_d=pp_refe_d; }
  choix_pplex=choix_direction(pt,pplex_noeud,pplex_g,pplex_d);

  /* TEST : on 'force' d'abord a avoir une reponse OUI pour imposer le choix
     correct sur : QUESTION, REQUETE, REPONSE, AUTRE. Puis on utilise la pplex
  if (NbChoixOui==0) ChoixMethode=CHOIX_ARBRE; else ChoixMethode=CHOIX_BEST; */

  if (ChoixMethode==CHOIX_ARBRE) gain_pplex=choix_arbre; else gain_pplex=choix_pplex;

  if (choix_arbre!=choix_pplex) { OkDirection=False; /*affiche_item(item[I_REFE],my_format); affiche_item(item[I_BEST],my_format);*/ }
  if (choix_arbre!=NULL) NbChoixOui++;
  
  if (Debug>=2)
   {
   printf("Noeud %d pplex=%.2lf - fg=%d pplex=%.2lf - fd=%d pplex=%.2lf\n",
	trouve_ML(pt),pplex_noeud,i_ML_gauche,pplex_g,i_ML_droite,pplex_d);
   printf("\tChoix du KS = %s\n",resu?"gauche":"droite");
   printf("\tChoix du ML = %s\n",gain_pplex?(gain_pplex==pt->fg?"gauche":"droite"):"NULL");
   }

  if (gain_pplex==NULL) return pt;
  else
   if (gain_pplex==pt->fg)
    {
    retour_noeud=parcoursML(pt->fg,item,newKS,pplex_g,niveau_max,niveau+1,indice_depart,my_format);
    free(newKS);
    RestorEtatNBEST(item,etat);
    return retour_noeud;
    }
   else
    {
    free(newKS);
    RestorEtatNBEST(item,etat);
    return parcoursML(pt->fd,item,tablKS,pplex_d,niveau_max,niveau+1,indice_depart,my_format);
    }
  }
 }
else ERREUR("Pourquoi est-on trop loin dans l'arbre ??","");
return NULL;
}

void meilleurML(ty_node pt, ty_item *item,double *bestpplex,ty_node *bestnode,ty_format my_format)
{
double pplex,pp_cnet,pp_refe;

if (pt)
 {
 pplex=calcule_pplex_nbest(item,trouve_ML(pt),&pp_cnet,&pp_refe,my_format);
 if (pplex<(*bestpplex)) { (*bestpplex)=pplex; (*bestnode)=pt; }
 meilleurML(pt->fg,item,bestpplex,bestnode,my_format);
 meilleurML(pt->fd,item,bestpplex,bestnode,my_format);
 }
}

/* on choisi les ML sur les exemples */

typedef struct
	{
	ty_node feuille;
	int nb;
	} type_freqfeuille;

int compar_freq_feuille(const void *a,const void *b)
{
type_freqfeuille *c,*d;
c=(type_freqfeuille*)a;
d=(type_freqfeuille*)b;
return d->nb-c->nb;
}

void choisit_ML(ty_node racine,int niveau_max,ty_format my_format)
{
int i,j,n,i_ML,nbbig=0,nb_diff_direction=0,nb_feuille,somm_feuille;
ty_node feuille;
type_freqfeuille tablfeuille[TAILLE_NBEST];
type_KS tablKS[TAILLE_NBEST];
double somme_pb=0,somme_pbref=0,pplex,pplex_ref,pplex_cnet;

NbFeuilleAtteinte=0;

for(n=0;n<NbNbest;n++)
  {
  if ((n+1)%500==0) fprintf(stderr,"\ten cours : test=%d\n",n+1);

  /*  Initialisation du KS  */
  for(i=0;i<TAILLE_NBEST;i++)
   {
   tablKS[i].gap_avant=tablKS[i].gap_apres=1; tablKS[i].ptKS=NULL;
   tablfeuille[i].feuille=NULL;
   tablfeuille[i].nb=0;
   }

  if ((i_ML=trouve_ML(racine))==-1) ERREUR("pas de ML a la racine ??","");
  pplex=calcule_pplex_nbest(TablNbest[n]->item,i_ML,&pplex_cnet,&pplex_ref,my_format);
  if (ChoixMethode==CHOIX_CNET) pplex=pplex_cnet; else
  if (ChoixMethode==CHOIX_REFE) pplex=pplex_ref;

  OkDirection=True; NbChoixOui=0;
 
  for(nb_feuille=0,i=I_BEST;i<I_BEST+1 /*TAILLE_NBEST*/;i++)
   if (TablNbest[n]->item[i])
    {
    feuille=parcoursML(racine,TablNbest[n]->item,tablKS,pplex,niveau_max,0,i,my_format);
    for(j=0;(j<nb_feuille)&&(tablfeuille[j].feuille!=feuille);j++);
    if (j==nb_feuille)
     {
     tablfeuille[j].feuille=feuille; tablfeuille[j].nb=1;
     nb_feuille++;
     if (nb_feuille==TAILLE_NBEST) ERREUR("TAILLE_NBEST trop petit","");
     }
    else tablfeuille[j].nb++;
    }

  qsort(tablfeuille,nb_feuille,sizeof(type_freqfeuille),compar_freq_feuille);
  /*for(j=0;j<nb_feuille;j++) fprintf(stderr,"%d] feuille %d\n",j+1,tablfeuille[j].nb);*/

  feuille=tablfeuille[0].feuille;

  if (OkDirection==False) nb_diff_direction++;

  /* on prends le meilleur ML, au sens de la perplex (ca marche pas) 
  feuille=NULL; pplex=MAX_PPLEX;
  meilleurML(racine,TablNbest[n]->item,&pplex,&feuille,my_format); */
  
  /*
  if ((i_ML=trouve_ML(feuille))==-1) ERREUR("Pas de ML possibles pour ",TablNbest[n]->ident);
  printf("%s\t%d\n",TablNbest[n]->ident,trouve_indice_noeud(i_ML));
  */

  printf("%s",TablNbest[n]->ident);
  for (j=somm_feuille=0;(j<nb_feuille)&&(((float)somm_feuille/(float)i)<1.0);j++)
   {
   somm_feuille+=tablfeuille[j].nb;
   feuille=tablfeuille[j].feuille;
   if ((i_ML=trouve_ML(feuille))==-1)
    {
    fprintf(stderr,"Pas de ML possibles pour %s - feuille=%s nb_feuille=%d => on met le modele general\n",
	TablNbest[n]->ident,feuille!=NULL?"PAS_NULL":"NULL",nb_feuille);
    i_ML=trouve_ML(racine);
    }
   printf("\t%d",trouve_indice_noeud(i_ML));
   }
  printf("\n");

  somme_pbref+=somm_proba(TablNbest[n]->item[I_REFE],trouve_ML(racine),&i,0,my_format);
  somme_pb+=somm_proba(TablNbest[n]->item[I_REFE],i_ML,&i,0,my_format);
  nbbig+=i;

  }

fprintf(stderr,"Resultats : methode=%d ",ChoixMethode);
if (niveau_max!=1000000) fprintf(stderr," niveau_max=%d - ",niveau_max);
fprintf(stderr,"OK_direction=%d/%d (%.2lf%%) - ",(NbNbest-nb_diff_direction),NbNbest,
	(double)((NbNbest-nb_diff_direction)*100)/(double)NbNbest);

pplex=exp(somme_pb/(double)nbbig*log(10.0));
pplex_ref=exp(somme_pbref/(double)nbbig*log(10.0));
fprintf(stderr,"PPLEX(general)=%.2lf - PPLEX(noeud)=%.2lf - gain=%.2lf (%.2lf%%)\n",
  pplex_ref,pplex,pplex_ref-pplex,(pplex_ref-pplex)*100/pplex_ref);
}

/*................................................................*/

/* charge TreeML avec les ML compiles et on les compile s'ils n'existent pas */

void charge_TreeML(ty_node pt,int *numnode,char *ch_arbre,char *ch_noeud_rep,ty_format my_format)
{
FILE *file;
static char ch[TailleLigne];
err_t retour;

if (pt)
 {
 (*numnode)++;
 if (TreeML[*numnode].i_ml!=-1) ERREUR("ML deja charge !!","");
 sprintf(ch,"%s/noeud_n%d.arpa",ch_noeud_rep,*numnode);
 if (file=fopen(ch,"rt"))
  {
  fclose(file);
  TreeML[*numnode].node=pt;
  sprintf(ch,"%s/noeud_n%d.arpa.desc",ch_noeud_rep,*numnode);
  if (!(file=fopen(ch,"rt")))
   {
   /* le ML n'est pas encore compile */
   sprintf(ch,"$LIA_TOOLTALN_REP/bin/compile_arpa -lexique %s.niv%d %s/noeud_n%d.arpa log_10 2g -dicho",
	ch_arbre,my_format->indice_ML+1,ch_noeud_rep,*numnode);
   system(ch);
   } else fclose(file);
  /* on le charge maintenant */
  sprintf(ch,"%s/noeud_n%d.arpa",ch_noeud_rep,*numnode);
  if ((retour=gram_module_init(ch,&(TreeML[*numnode].i_ml),2,0))!=CORRECT)
   {
   fprintf(stderr,"retour=%d\n",retour);
   ERREUR("Mauvaise lecture du modele :",ch);
   }
  if (*numnode==0) I_ModeleGeneral=TreeML[*numnode].i_ml;
  }
 else /*fprintf(stderr,"Warning : le fichier %s n'existe pas\n",ch)*/;
 charge_TreeML(pt->fg,numnode,ch_arbre,ch_noeud_rep,my_format);
 charge_TreeML(pt->fd,numnode,ch_arbre,ch_noeud_rep,my_format);
 }
}

/*................................................................*/

ty_format lis_descripteur(char *fich_desc,ty_node *racine,char *nom_arbre,char *nom_noeud_rep)
{
FILE *file;
char ch[TailleLigne],*pt,ch2[TailleLigne];
int si_arbre=0,si_noeud_rep=0;
ty_format my_format;

if (!(file=fopen(fich_desc,"rt"))) { fprintf(stderr,"can't open : %s\n",fich_desc); exit(0); }
while(fgets(ch,TailleLigne,file))
 {
 if (!strncmp(ch,"ARBRE",5))
  {
  si_arbre=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  sprintf(ch2,"%s.desc_lex",pt);
  my_format=LectureDescripteurFormat(ch2);
  (*racine)=LectureTree(pt,my_format);
  strcpy(nom_arbre,pt);
  } else
 if (!strncmp(ch,"NOEUD_REP",9))
  {
  si_noeud_rep=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(nom_noeud_rep,pt);
  }
 }
if (!si_arbre) ERREUR("il manque 'ARBRE' dans ",fich_desc);
if (!si_noeud_rep) ERREUR("il manque 'NOEUD_REP' dans ",fich_desc);
fclose(file);
return my_format;
}

void main(argc,argv)
 int argc;
 char **argv;
{
ty_node racine;
char *choix,*fich_corpus,nom_arbre[TailleLigne],nom_noeud_rep[TailleLigne];
int n,i,niveau_max=1000000;
ty_format my_format;

if ((argc==1)||(!strcmp(argv[1],"-h")))
 {
 fprintf(stderr,"Syntaxe : %s [-h] -desc <fichier descripteur> -nbest <fich reco nbest> \
 [-niv [0-??]] [-choix [1 2 3 4]]\n",argv[0]);
 exit(0);
 }

fprintf(stderr,"Chargement de l'arbre\n");
for (n=1;n<argc;n++)
 if (!strcmp(argv[n],"-desc")) my_format=lis_descripteur(argv[++n],&racine,nom_arbre,nom_noeud_rep); else
 if (!strcmp(argv[n],"-nbest"))
  {
  NbItemref=0;
  ArbreNbest=charge_itemref(argv[++n],my_format);
  TablNbest=(ty_itemref*)malloc(sizeof(ty_itemref)*NbItemref);
  NbNbest=NbItemref;
  range_itemref_tabl(ArbreNbest,TablNbest);
  }
 else
 if (!strcmp(argv[n],"-niv")) sscanf(argv[++n],"%d",&niveau_max);
 else
 if (!strcmp(argv[n],"-choix")) sscanf(argv[++n],"%d",&ChoixMethode);
 else
  {
  fprintf(stderr,"ERREUR : je ne connais pas l'option '%s' ....\n",argv[n]);
  exit(0);
  }
fprintf(stderr,"Termine\n");

n=-1;
fprintf(stderr,"Chargement des modeles\n");
charge_TreeML(racine,&n,nom_arbre,nom_noeud_rep,my_format);
fprintf(stderr,"\t-> on a lu %d ML\n",n);
fprintf(stderr,"Termine\n");
choisit_ML(racine,niveau_max,my_format);

}
 
