/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  MODIF HMIHY 0302 : on arrete des que le mot est XX (end of significant word)  */

/*  lia_sct_pplex : gestion de la perplexite (ngram)  */

#include <lia_sct.h>
#include <lia_sct_pplex.h>

/*................................................................*/

/* gestion des ngram pour la perplexite - on prevoit jusqu'au 2grams */

/*  On a 3 arbres globaux contenant les modeles ngrams :
    - le premier appele : RacineOuiGram correspondant aux item ayant repondu oui
    - le deuxieme       : RacineNonGram correspondant aux item ayant repondu non
    - le troisieme      : RacineAllGram groupant les 2 corpus pour evaluer la perplexite sur le noeud pere
    les 3 arbres ont une structure identique, c'est juste leur remplissage qui differe  */

ty_niveau RacineOuiGram,RacineNonGram,RacineAllGram;

/* Gestion de base */

ty_niveau nouveau_niveau(gram,nb)
 wrd_index_t gram;
 long nb;
{
ty_niveau pt;
pt=(ty_niveau)my_malloc(sizeof(struct type_niveau));
pt->gram=gram;
pt->nb=nb;
pt->pb=0.;
pt->next=pt->fg=pt->fd=NULL;
return pt;
}

ty_niveau ajoute_niveau2(gram2,racine)
 wrd_index_t gram2;
 ty_niveau racine;
{
if (racine==NULL) racine=nouveau_niveau(gram2,1);
else
 if (racine->gram==gram2) racine->nb++;
 else
  if (racine->gram>gram2)
   racine->fg=ajoute_niveau2(gram2,racine->fg);
  else
   racine->fd=ajoute_niveau2(gram2,racine->fd);
return racine;
}

ty_niveau ajoute_niveau1(wrd_index_t gram1,wrd_index_t gram2,ty_niveau racine,long * nb,ty_format my_format)
{
/*
fprintf(stderr,"on ajoute : %s , %s\n",
  renvoi_chaine(gram1,my_format->indice_ML,my_format),renvoi_chaine(gram2,my_format->indice_ML,my_format));
*/

if (racine==NULL)
 {
 (*nb)++;
 racine=nouveau_niveau(gram1,1);
 racine->next=ajoute_niveau2(gram2,racine->next);
 }
else
 if (racine->gram==gram1)
  {
  (*nb)++;
  racine->nb++;
  racine->next=ajoute_niveau2(gram2,racine->next);
  }
 else
  if (racine->gram>gram1)
   racine->fg=ajoute_niveau1(gram1,gram2,racine->fg,nb,my_format);
  else
   racine->fd=ajoute_niveau1(gram1,gram2,racine->fd,nb,my_format);
return racine;
}

/*................................................................*/

/* fonction de remplissage de l'arbre cree : compte et proba */

ty_niveau ajoute_item_ngram(ty_item ph,ty_niveau racine,long *nb,ty_format my_format)
{
/* on entre les 2grams - MODIF HMIHY 0302 : on arrete des que le mot est XX (end of significant word) */
for(;(ph)&&(ph->suiv)&&(strcmp(renvoi_chaine(ph->code[my_format->indice_ML],my_format->indice_ML,my_format),"XX"));ph=ph->suiv)
 racine=ajoute_niveau1(ph->code[my_format->indice_ML],ph->suiv->code[my_format->indice_ML],racine,nb,my_format);
return racine;
}

void rempli_proba(nb,racine,cutoff)
 long nb;
 ty_niveau racine;
{
if (racine)
 {
 if ((racine->nb<=cutoff)||(nb==0)) racine->pb=-PENAL_UNK;
 else racine->pb=log10((double)racine->nb/(double)nb);
 rempli_proba(racine->nb,racine->next,CUTOFF_2);
 rempli_proba(nb,racine->fg,cutoff);
 rempli_proba(nb,racine->fd,cutoff);
 }
}

void raz_gram(racine) /* remet compte et proba a zero */
 ty_niveau racine;
{
if (racine)
 {
 racine->nb=(long)0;
 racine->pb=(double)0;
 raz_gram(racine->next);
 raz_gram(racine->fg);
 raz_gram(racine->fd);
 }
}

/*................................................................*/

/* Calcul des probas et de la perplexite */

ty_niveau recherche_gram(gram,racine)
 wrd_index_t gram;
 ty_niveau racine;
{
if ((racine==NULL)||(racine->gram==gram)) return racine;
if (racine->gram>gram) return recherche_gram(gram,racine->fg);
else return recherche_gram(gram,racine->fd);
}

int proba_1_gram(pb,gram1,racine_ngram)
 double *pb;
 wrd_index_t gram1;
 ty_niveau racine_ngram;
{
ty_niveau pt1;
pt1=recherche_gram(gram1,racine_ngram);
if (pt1==NULL) return 0;
if (pt1->nb>CUTOFF_1) *pb=-pt1->pb; else return 0;
return 1;
}

int proba_2_gram(pb,gram1,gram2,racine_ngram)
 double *pb;
 wrd_index_t gram1,gram2;
 ty_niveau racine_ngram;
{
ty_niveau pt1,pt2;

pt1=recherche_gram(gram1,racine_ngram);

if ((pt1==NULL)||(pt1->nb<=CUTOFF_2)) return 0;

*pb=-pt1->pb;

pt2=recherche_gram(gram2,pt1->next);

if ((pt2==NULL)||(pt2->nb<=CUTOFF_2)) (*pb)+=PENAL_1;
else *pb=-pt2->pb;

return 1;
}

double somme_proba_2g(pt1,racine,nb,racine_ngram)
 ty_niveau pt1,racine,racine_ngram;
 long *nb;
{
int choix;
double total;
if (racine)
 {
 if (pt1==NULL) choix=1; else choix=2;
 switch (choix)
  {
  case 1 : total=somme_proba_2g(racine,racine->next,nb,racine_ngram); break;
  case 2 :
	 if (racine->nb>CUTOFF_2) total=-racine->pb;
	 else if (!proba_1_gram(&total,racine->gram,racine_ngram)) total=10;
	 total*=racine->nb;
	 (*nb)+=racine->nb;
	 break;
  }
 total+=somme_proba_2g(pt1,racine->fg,nb,racine_ngram);
 total+=somme_proba_2g(pt1,racine->fd,nb,racine_ngram);
 return total;
 }
return 0;
}

double calcule_perplexity(ty_niveau racine)
{
long nb;
double total;
nb=0;
total=somme_proba_2g(NULL,racine,&nb,racine);
/*printf("nb=%ld total=%lf\n",nb,total);*/
return exp(total/(double)nb*log(10.0));
}

double calcule_perplexity_test(ty_niveau racine,ty_sample corpus,int siaf,ty_format my_format)
{
ty_sample pt_sample;
ty_item pt_ph;
long nb;
double total,pb;

if (siaf) printf("Affichage des item comptant pour le calcul de perplexite :\n");

for(nb=0,total=0,pt_sample=corpus;pt_sample;pt_sample=pt_sample->suiv)
 {
 if (siaf) affiche_item(pt_sample->item,my_format);

 /* MODIF HMIHY 0302 : on arrete des que le mot est XX (end of significant word) */
 for(pt_ph=pt_sample->item;(pt_ph->suiv)&&
     (strcmp(renvoi_chaine(pt_ph->suiv->code[my_format->indice_ML],my_format->indice_ML,my_format),"XX"));pt_ph=pt_ph->suiv)
  {
  if (!proba_2_gram(&pb,pt_ph->code[my_format->indice_ML],pt_ph->suiv->code[my_format->indice_ML],racine)) pb=PENAL_UNK;

  if (siaf) 
   printf("TEST : P(%s,%s)=%.4lf\n",renvoi_chaine(pt_ph->code[my_format->indice_ML],my_format->indice_ML,my_format),
    renvoi_chaine(pt_ph->suiv->code[my_format->indice_ML],my_format->indice_ML,my_format),pb);
  
 
  total+=pb;
  nb++;
  }
 }
return exp(total/(double)nb*log(10.0));
}

double calcule_perplexity_test_succes(char booleen,ty_niveau racine,ty_sample *corpus,
				      type_succes *bilan,int nb_sample,int siaf,ty_format my_format)
{
ty_item pt_ph;
long nb;
double total,pb;
int i;

if (siaf) printf("Affichage des item comptant pour le calcul de perplexite :\n");

for(nb=0,total=0,i=0;i<nb_sample;i++)
 {
 if (bilan[i].good==booleen)
 /* MODIF HMIHY 0302 : on arrete des que le mot est XX (end of significant word) */
  for(pt_ph=corpus[i]->item;(pt_ph->suiv)&&
     (strcmp(renvoi_chaine(pt_ph->suiv->code[my_format->indice_ML],my_format->indice_ML,my_format),"XX"));pt_ph=pt_ph->suiv)
   {
   if (!proba_2_gram(&pb,pt_ph->code[my_format->indice_ML],pt_ph->suiv->code[my_format->indice_ML],racine)) pb=PENAL_UNK;
   total+=pb;
   nb++;
   }
  }
return exp(total/(double)nb*log(10.0));
}

void affiche_proba(char *ch1,char *ch2,char *ch3,ty_niveau racine,ty_format my_format)
{
int choix;

if (racine)
 {
 if (ch1[0]=='\0') choix=1; else if (ch2[0]=='\0') choix=2; else choix=3;
 switch (choix)
  {
  case 1 : strcpy(ch1,renvoi_chaine(racine->gram,my_format->indice_ML,my_format)); break;
  case 2 : strcpy(ch2,renvoi_chaine(racine->gram,my_format->indice_ML,my_format)); break;
  case 3 : strcpy(ch3,renvoi_chaine(racine->gram,my_format->indice_ML,my_format)); break;
  }

 printf("Proba : %s , %s , %s = %.2lf  (nb=%ld)\n",ch1,ch2,ch3,racine->pb,racine->nb);

 affiche_proba(ch1,ch2,ch3,racine->next,my_format);

 switch (choix)
  {
  case 1 : ch1[0]='\0';
  case 2 : ch2[0]='\0';
  case 3 : ch3[0]='\0';
  }

 affiche_proba(ch1,ch2,ch3,racine->fg,my_format);
 affiche_proba(ch1,ch2,ch3,racine->fd,my_format);
 }
}

/*................................................................*/

ty_niveau rempli_ngram(ty_item pt,ty_niveau racine,long *nb,ty_format my_format)
{
/* on entre les 2grams - MODIF HMIHY 0302 : on arrete des que le mot est XX (end of significant word) */
for(pt=pt->suiv;(pt)&&(strcmp(renvoi_chaine(pt->code[my_format->indice_ML],my_format->indice_ML,my_format),"XX"));pt=pt->suiv)
 racine=ajoute_niveau1(pt->pred->code[my_format->indice_ML],pt->code[my_format->indice_ML],racine,nb,my_format);
return racine;
}

/*................................................................*/

/* parcours une liste d'item pour remplir les niveaux d'un ngram */

ty_niveau rempli_niveaux(ty_sample corpus,ty_format my_format)
{
ty_niveau racine=NULL;
long nb_gram=0;
ty_sample pt;

for (pt=corpus;pt;pt=pt->suiv) racine=rempli_ngram(pt->item,racine,&nb_gram,my_format);
rempli_proba(nb_gram,racine,0);
/*  test de coherence : ca a l'air d'aller !!
{
char ch1[TailleLigne],ch2[TailleLigne],ch3[TailleLigne];
ch1[0]=ch2[0]=ch3[0]='\0';
affiche_proba(ch1,ch2,ch3,racine,my_format);
}
*/
return racine;
}
  
