/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*  Construit les ressources necessaires au toolkit 'LIA_SCT'
    en fonction d'un fichier descripteur  */
/*  FRED 0600  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <lia_tree_avl.h>

#define TailleLigne	8000

void ERREUR(char *ch1,char *ch2)
{
fprintf(stderr,"ERREUR : %s %s\n",ch1,ch2);
exit(0);
}

#define max(a,b)	((a)>(b)?(a):(b))
#define min(b,a)	((a)>(b)?(a):(b))

/* exemple de fichier descripteur :

# francaise phonetisee au LIA et decrite sur 2 niveaux :
# phoneme et cohorte
#
# FRED 0600

NB_NIVEAUX      2
INDICE_ML       1
CORPUS_TRAIN    ./DATA/list_localite.train.fmt
CORPUS_TEST     ./DATA/list_localite.test.fmt
CORPUS_DEV      NULL
ARBRE           ./DATA/list_localite.arbre
FEUILLE_REP     ./FEUILLE
TEST_REP        ./TEST
DEVT_REP	./DEVT
NOEUD_REP       ./NOEUD
SCRIPT_REP      ./SCRIPT
TAILLE_FEUILLE  2000
*/

/* variables globales pour la description des donnees */

/* on ne limite pas la taille de l'arbre */
#define PROF_MAXI	10000

#define MAX_NIVEAUX	1000 /* nb max de niveau pour representer un item (pour NIVEAUX_QUESTION) */

int TablNiveauxQuestion[MAX_NIVEAUX];

int NB_NIVEAUX,INDICE_ML,TAILLE_FEUILLE,NB_MAX_GLB_QUEST,IF_GLOBAL_QUESTION;
char CORPUS_TRAIN[TailleLigne],CORPUS_TEST[TailleLigne],CORPUS_DEV[TailleLigne],
     ARBRE[TailleLigne],FEUILLE_REP[TailleLigne],TEST_REP[TailleLigne],DEVT_REP[TailleLigne],
     NOEUD_REP[TailleLigne],SCRIPT_REP[TailleLigne],FICH_DESCRIPTEUR[TailleLigne],MODE[TailleLigne],
     CRITERE[TailleLigne],DELTA[TailleLigne];

/*................................................................*/

/* creation des fichiers lexiques pour les references et chaque niveaux */

void rempli_arbre(lia_avl_t *tabl,char *nomfic)
{
int nb,i;
char ch[TailleLigne*10],*pt,*pafini;
FILE *file;

/* on rempli les arbres */
if (!(file=fopen(nomfic,"rt"))) ERREUR("can't open :",nomfic);

for (nb=1,pafini=fgets(ch,TailleLigne*10,file);pafini;nb++)
 {
 if (ch[0]=='@') { fprintf(stderr,"ERREUR1 : ligne %d incorrect dans le fichier %s\n",nb,nomfic); exit(0); }
 pt=strtok(ch," \t\n");
 if (pt) tabl[NB_NIVEAUX]=lia_ajoute_element_avl(tabl[NB_NIVEAUX],pt,True,NULL);
 if (pt) pt=strtok(NULL," \t\n");
 if (pt==NULL) { fprintf(stderr,"ERREUR2 : ligne %d incorrect dans le fichier %s\n",nb,nomfic); exit(0); }
 while (pt)
  {
	  /*fprintf(stderr,"DEBUG: %s\n",pt);*/

  for(i=0;i<NB_NIVEAUX;i++)
   {
   if (pt==NULL) { fprintf(stderr,"ERREUR3 : ligne %d incorrect dans le fichier %s (NB_NIVEAUX=%d)\n",nb,nomfic,NB_NIVEAUX); exit(0); }
   tabl[i]=lia_ajoute_element_avl(tabl[i],pt,True,NULL);
   pt=strtok(NULL," \t\n");
   }
  }
 pafini=fgets(ch,TailleLigne*10,file);
 if (pafini)
  {
  if (ch[0]=='&')	/* global questions */
   {
   if (IF_GLOBAL_QUESTION)
    {
    for(pt=strtok(ch+2," \t\n"),i=0;pt;pt=strtok(NULL," \t\n"),i++)
     tabl[NB_NIVEAUX+2]=lia_ajoute_element_avl(tabl[NB_NIVEAUX+2],pt,True,NULL);
    NB_MAX_GLB_QUEST=max(NB_MAX_GLB_QUEST,i);
    }
   pafini=fgets(ch,TailleLigne*10,file);
   }
  if (!strcmp(MODE,"SUPERVISE"))
   {
   if ((pafini==NULL)||(ch[0]!='@'))
    { fprintf(stderr,"ERREUR4 : ligne %d incorrect dans le fichier %s\n",nb,nomfic); exit(0); }
   nb++;
   pt=strtok(ch," \t\n");
   pt=strtok(NULL," \t\n");
   tabl[NB_NIVEAUX+1]=lia_ajoute_element_avl(tabl[NB_NIVEAUX+1],pt,True,NULL);
   pafini=fgets(ch,TailleLigne*10,file);
   }
  }
 }
fclose(file);
}

void sauve_arbre(lia_avl_t pt,FILE *file,int *num)
{
if (pt==NULL) return;
sauve_arbre(pt->fg,file,num);
fprintf(file,"%d\t%s\n",(*num)++,pt->info);
sauve_arbre(pt->fd,file,num);
}

void ecris_question_niveau(lia_avl_t pt,FILE *file,int niveau)
{
if (pt==NULL) return;
ecris_question_niveau(pt->fg,file,niveau);
/* on ne fait une question que des items qui apparaissent au moins TAILLE_FEUILLE dans le corpus */
if (pt->nb>=TAILLE_FEUILLE)
 {
 if (!strcmp(pt->info,"XX")) fprintf(stderr,"WARNING: the symbol 'XX' is discared from the list of possible questions\n");
 else fprintf(file,"%s\t%d\n",pt->info,niveau);
 }
ecris_question_niveau(pt->fd,file,niveau);
}

void creation_fichiers_lexique()
{
lia_avl_t *tabl;
int i,nb;
FILE *file;
char ch[TailleLigne];

fprintf(stderr,"\nCreation des fichiers lexiques : \n");

/* on cree le tableau d'arbres correspondant aux niveaux  :
	case NB_NIVEAUX = reference
	case NB_NIVEAUX+1 = etiquette
	case NB_NIVEAUX+2 = all the global questions */

tabl=(lia_avl_t *)malloc(sizeof(lia_avl_t)*(NB_NIVEAUX+2+1));
for(i=0;i<NB_NIVEAUX+2+1;i++) tabl[i]=NULL;

rempli_arbre(tabl,CORPUS_TRAIN);
/* rempli_arbre(tabl,CORPUS_TEST); ON NE RAJOUTE PLUS LE LEXIQUE DU TEST */
if (strcmp(CORPUS_DEV,"NULL")) rempli_arbre(tabl,CORPUS_DEV);

/* on sauvegarde les fichiers */

sprintf(ch,"%s.ref",ARBRE);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"0\t<UNK>\n");
nb=1; sauve_arbre(tabl[NB_NIVEAUX],file,&nb);
fclose(file);

if (!strcmp(MODE,"SUPERVISE"))
 {
 sprintf(ch,"%s.eti",ARBRE);
 if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
 nb=1; sauve_arbre(tabl[NB_NIVEAUX+1],file,&nb);
 fclose(file);
 }

for(i=0;i<NB_NIVEAUX;i++)
 {
 sprintf(ch,"%s.niv%d",ARBRE,i+1);
 if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
 fprintf(file,"0\t<UNK>\n");
 nb=1; sauve_arbre(tabl[i],file,&nb);
 fclose(file);
 }

if (IF_GLOBAL_QUESTION)
 {
 sprintf(ch,"%s.glb_quest",ARBRE);
 if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
 nb=1; sauve_arbre(tabl[NB_NIVEAUX+2],file,&nb);
 fclose(file);
 }

/* compilation des lexiques */
fprintf(stderr,"\tcompilation de : %s.ref\n",ARBRE);
sprintf(ch,"$LIA_SCT/bin/lia_compile_lexique %s.ref",ARBRE); system(ch);
fprintf(stderr,"\tcompilation de : %s.eti\n",ARBRE);
sprintf(ch,"$LIA_SCT/bin/lia_compile_lexique %s.eti",ARBRE); system(ch);
for(i=0;i<NB_NIVEAUX;i++)
 {
 fprintf(stderr,"\tcompilation de : %s.niv%d\n",ARBRE,i+1);
 sprintf(ch,"$LIA_SCT/bin/lia_compile_lexique %s.niv%d",ARBRE,i+1); system(ch);
 }
if (IF_GLOBAL_QUESTION)
 {
 fprintf(stderr,"\tcompilation de : %s.glb_quest",ARBRE);
 sprintf(ch,"$LIA_SCT/bin/lia_compile_lexique %s.glb_quest",ARBRE); system(ch);
 }
fprintf(stderr,"\n\nCreation de la liste de questions locales : \n");
sprintf(ch,"%s.question",ARBRE);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
/* on n'ecrit que les questions pour lesquelles le niveau a ete selectionne */
for(i=1;i<=NB_NIVEAUX;i++)
 if (TablNiveauxQuestion[i]) ecris_question_niveau(tabl[i-1],file,i);
fprintf(stderr,"\tecriture du fichier %s.question\n",ARBRE);
fclose(file);

fprintf(stderr,"\nCreation du descripteur de lexique : \n");
sprintf(ch,"%s.desc_lex",ARBRE);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"# Descripteur de fichier pour l'arbre : %s\n#\n",ARBRE);
fprintf(file,"MODE %s\nCRITERE %s\n",MODE,CRITERE);
fprintf(file,"DELTA %s\n",DELTA);
fprintf(file,"NB_NIVEAUX %d\nINDICE_ML %d\nNB_MAX_GLB_QUEST %d\n",NB_NIVEAUX,INDICE_ML,NB_MAX_GLB_QUEST);
fprintf(file,"LEXIQUE REF %s.ref\n",ARBRE);
if (!strcmp(MODE,"SUPERVISE")) fprintf(file,"LEXIQUE ETI %s.eti\n",ARBRE);

for(i=0;i<NB_NIVEAUX;i++)
 fprintf(file,"LEXIQUE %d %s.niv%d\n",i+1,ARBRE,i+1);
if (NB_MAX_GLB_QUEST>0) fprintf(file,"GLOBAL_QUEST %s.glb_quest\n",ARBRE);
fprintf(file,"\n#\n# Termine\n");
fclose(file);
fprintf(stderr,"\tTermine\n");
}

/*................................................................*/

/* ecriture des fichiers script */

void creation_fichiers_scripts()
{
FILE *file;
char ch[TailleLigne];

fprintf(stderr,"\nCreation des fichiers scripts dans le repertoire : %s\n",SCRIPT_REP);

sprintf(ch,"%s/grow_tree",SCRIPT_REP);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"#!/usr/bin/csh\n$LIA_SCT_REP/bin/lia_sct_build %s\n",FICH_DESCRIPTEUR);
fclose(file);
/*
sprintf(ch,"%s/build_lm",SCRIPT_REP);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"#!/usr/bin/csh\nrm %s/feuille*\nrm %s/feuille*\nrm %s/feuille*\nrm %s/noeud*\n\n",
	FEUILLE_REP,TEST_REP,DEVT_REP,NOEUD_REP);
fprintf(file,"$LIA_SCT_REP/bin/lia_sct_use -desc %s -corpus_feuille %s < %s\n",
 FICH_DESCRIPTEUR,FEUILLE_REP,CORPUS_TRAIN);
fprintf(file,"$LIA_SCT_REP/bin/lia_sct_use -desc %s -corpus_feuille %s < %s\n",
 FICH_DESCRIPTEUR,TEST_REP,CORPUS_TEST);
if (strcmp(CORPUS_DEV,"NULL"))
 fprintf(file,"$LIA_SCT_REP/bin/lia_sct_use -desc %s -corpus_feuille %s < %s\n",
  FICH_DESCRIPTEUR,DEVT_REP,CORPUS_DEV);
fprintf(file,"$LIA_SCT_REP/bin/lia_sct_ml -desc %s >! %s.log_ml\n",FICH_DESCRIPTEUR,ARBRE);
fprintf(file,"$LIA_SCT_REP/bin/analyse_sortie < %s.log_ml\n",ARBRE);
fclose(file);
*/
sprintf(ch,"%s/print_tree",SCRIPT_REP);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"#!/usr/bin/csh\n");
fprintf(file,"$LIA_SCT_REP/bin/lia_sct_use -desc %s -affiche\n",
 FICH_DESCRIPTEUR);
fclose(file);

/*
sprintf(ch,"%s/do_compile",SCRIPT_REP);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"#!/usr/bin/csh\n");
fprintf(file,"$LIA_SCT_REP/script/lance_fprobs %s %s\n",ARBRE,NOEUD_REP);
fprintf(file,"$LIA_SCT_REP/bin/mixture_modele %s -compile\n",FICH_DESCRIPTEUR);
fclose(file);

sprintf(ch,"%s/lance_tout",SCRIPT_REP);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"#!/usr/bin/csh\n");
fprintf(file,"csh %s/grow_tree\ncsh %s/build_lm\ncsh %s/run_test\ncsh %s/do_compile\n",
	SCRIPT_REP,SCRIPT_REP,SCRIPT_REP,SCRIPT_REP);
fclose(file);
*/
fprintf(stderr,"\tTermine\n");

/*
sprintf(ch,"%s.varenv",ARBRE);
if (!(file=fopen(ch,"wt"))) ERREUR("can't write in :",ch);
fprintf(file,"# Variables d'environnement correspondant a l'arbre : %s\n#\n",ARBRE);
fprintf(file,"setenv DICO_REFERENCE %s.ref\n",ARBRE);
fprintf(file,"setenv DESCRIPTEUR %s.desc_lex\n",ARBRE);
fprintf(file,"setenv QUESTION %s.question\n",ARBRE);
fprintf(file,"setenv CORPUS_TRAIN %s\n",CORPUS_TRAIN);
fprintf(file,"setenv CORPUS_TEST %s\n",CORPUS_TEST);
fprintf(file,"setenv CORPUS_DEV %s\n",CORPUS_DEV);
fprintf(file,"setenv ARBRE %s\n",ARBRE);
fprintf(file,"setenv PROF_MAXI %d\n",PROF_MAXI);
fprintf(file,"setenv TAILLE_FEUILLE %d\n",TAILLE_FEUILLE);
fprintf(file,"setenv FEUILLE_REP %s\n",FEUILLE_REP);
fprintf(file,"setenv FEUILLE $FEUILLE_REP/feuille\n");
fprintf(file,"setenv NOEUD_REP %s\n",NOEUD_REP);
fprintf(file,"setenv NOEUD $NOEUD_REP/noeud\n");
fprintf(file,"setenv TEST_REP %s\n",TEST_REP);
fprintf(file,"setenv TEST $TEST_REP/test\n");
fprintf(file,"#\n#  Termine\n");
fclose(file);
fprintf(stderr,"\tTermine\n");
*/
}

/*................................................................*/

int main(int argc, char **argv)
{
char ch[TailleLigne],*pt;
int nb,i,si_niveaux_question=0;
FILE *file;

if (argc<2)
 {
 fprintf(stderr,"Syntaxe : %s <fichier descripteur>\n",argv[0]);
 exit(0);
 }

if (!(file=fopen(argv[1],"rt"))) ERREUR("can't open ",argv[1]);

/* init */

for (nb=0;nb<MAX_NIVEAUX;nb++) TablNiveauxQuestion[nb]=0;

strcpy(FICH_DESCRIPTEUR,argv[1]);

NB_NIVEAUX=INDICE_ML=TAILLE_FEUILLE=NB_MAX_GLB_QUEST=0;
CORPUS_TRAIN[0]=CORPUS_TEST[0]=CORPUS_DEV[0]=ARBRE[0]='\0';
FEUILLE_REP[0]=TEST_REP[0]=DEVT_REP[0]=NOEUD_REP[0]=SCRIPT_REP[0]='\0';
MODE[0]=CRITERE[0]=DELTA[0]='\0';
IF_GLOBAL_QUESTION=False;

for(nb=0;fgets(ch,TailleLigne,file);nb++)
 {
 if (!strncmp(ch,"NB_NIVEAUX",10))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  sscanf(pt,"%d",&NB_NIVEAUX);
  } else
 if (!strncmp(ch,"GLOBAL_QUESTION",15))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (!strcmp(pt,"TRUE")) IF_GLOBAL_QUESTION=True;
  } else
 if (!strncmp(ch,"NIVEAUX_QUESTION",16))
  {
  pt=strtok(ch," \t\n");
  for(pt=strtok(NULL," \t\n");pt;pt=strtok(NULL," \t\n"))
   {
   sscanf(pt,"%d",&i);
   TablNiveauxQuestion[i]=1;
   }
  si_niveaux_question=1;
  } else
 if (!strncmp(ch,"INDICE_ML",9))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  sscanf(pt,"%d",&INDICE_ML);
  } else
 if (!strncmp(ch,"TAILLE_FEUILLE",14))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  sscanf(pt,"%d",&TAILLE_FEUILLE);
  } else
 if (!strncmp(ch,"CORPUS_TRAIN",12))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(CORPUS_TRAIN,pt);
  } else
 if (!strncmp(ch,"CORPUS_TEST",11))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(CORPUS_TEST,pt);
  } else
 if (!strncmp(ch,"CORPUS_DEV",10))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(CORPUS_DEV,pt);
  } else
 if (!strncmp(ch,"ARBRE",5))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(ARBRE,pt);
  } else
 if (!strncmp(ch,"FEUILLE_REP",11))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(FEUILLE_REP,pt);
  } else
 if (!strncmp(ch,"DEVT_REP",8))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(DEVT_REP,pt);
  } else
 if (!strncmp(ch,"TEST_REP",8))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(TEST_REP,pt);
  } else
 if (!strncmp(ch,"NOEUD_REP",9))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(NOEUD_REP,pt);
  } else
 if (!strncmp(ch,"SCRIPT_REP",10))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(SCRIPT_REP,pt);
  } else
 if (!strncmp(ch,"MODE",4))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(MODE,pt);
  } else
 if (!strncmp(ch,"CRITERE",7))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(CRITERE,pt);
  } else
 if (!strncmp(ch,"DELTA",5))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(DELTA,pt);
  }
 }

fclose(file);

/* valeur par defaut */
if (INDICE_ML==0) INDICE_ML=1;
if (FEUILLE_REP[0]==0) strcpy(FEUILLE_REP,"NULL");
if (TEST_REP[0]==0) strcpy(TEST_REP,"NULL");
if (DEVT_REP[0]==0) strcpy(DEVT_REP,"NULL");
if (NOEUD_REP[0]==0) strcpy(NOEUD_REP,"NULL");
if (CORPUS_TEST[0]==0) strcpy(CORPUS_TEST,"NULL");
if (CORPUS_DEV[0]==0) strcpy(CORPUS_DEV,"NULL");

/* valeur obligatoire */
if (NB_NIVEAUX==0) ERREUR("NB_NIVEAUX non defini","");
if (TAILLE_FEUILLE==0) ERREUR("TAILLE_FEUILLE non defini","");
if (CORPUS_TRAIN[0]==0) ERREUR("CORPUS_TRAIN non defini","");
if (ARBRE[0]==0) ERREUR("ARBRE non defini","");
/*if (SCRIPT_REP[0]==0) ERREUR("SCRIPT_REP non defini","");*/
if (MODE[0]==0) ERREUR("MODE non defini","");
if (CRITERE[0]==0) ERREUR("CRITERE non defini","");
if (DELTA[0]==0) ERREUR("DELTA non defini","");

if (si_niveaux_question==0) ERREUR("NIVEAUX_QUESTION non defini","");

creation_fichiers_lexique();
/*creation_fichiers_scripts();*/

}
  
