/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*  Merge un ARPA calcule avec le CMU avec des probas
    bigramme calcule en local (pour eviter le bug CMU)
    du Good Turing avec les singletons a -99.99.
    Ne modifie QUE les probas a -99  */
/*  FRED 0301  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <lia_tree_avl.h>

/*................................................................*/

#define TailleLigne	800

#define True	1
#define False	0

void ERREUR(char *ch1,char *ch2)
{
fprintf(stderr,"ERREUR : %s %s\n",ch1,ch2);
exit(0);
}

/*................................................................*/


/*  calcule des probas simples 2g sans backoff sur
    un fichier corpus avec 1 mot par ligne  */

lia_avl_t Tree1g=NULL,Tree2g=NULL;

void affiche_proba_arbre(lia_avl_t racine,lia_avl_t tree1g)
{
static char graf1[TailleLigne];
int i,nb;
double pb;

if (racine)
 {
 for(i=0;(racine->info[i])&&(racine->info[i]!=' ');i++) graf1[i]=racine->info[i];
 if (racine->info[i]!=' ') ERREUR("mauvais format dans :",racine->info);
 graf1[i]='\0';
 if (!(lia_recherche_avl(tree1g,graf1,&nb))) ERREUR("je ne trouve pas :",graf1);
 if (nb<racine->nb) nb=racine->nb;
 pb=log10((double)racine->nb/((double)nb+1));
 if (pb==(double)0) { fprintf(stderr,"big=%s nb(big)=%d nb(%s)=%d\n",racine->info,racine->nb,graf1,nb); exit(0); }
 printf("%lf %s\n",pb,racine->info);
 affiche_proba_arbre(racine->fg,tree1g);
 affiche_proba_arbre(racine->fd,tree1g);
 }
}

double calcule_proba(char *graf1,char *graf2)
{
static char bigramme[TailleLigne];
int nbbig,nb1g;

sprintf(bigramme,"%s %s",graf1,graf2);
if (!(lia_recherche_avl(Tree2g,bigramme,&nbbig))) ERREUR("je ne trouve pas :",bigramme);
if (!(lia_recherche_avl(Tree1g,graf1,&nb1g))) ERREUR("je ne trouve pas :",graf1);
if (nb1g<nbbig) nb1g=nbbig;
return log10((double)nb1g/((double)nb1g+1));
}

/*................................................................*/

void construit_arbre_gram(char *nomfich)
{
char ch[TailleLigne],graf1[TailleLigne],*graf2,bigramme[TailleLigne];
int nb;
FILE *file;

if (!(file=fopen(nomfich,"rt"))) ERREUR("can't open :",nomfich);
fgets(ch,TailleLigne,file);
graf2=strtok(ch," \t\n");
strcpy(graf1,graf2);
for(nb=0;fgets(ch,TailleLigne,file);nb++)
 {
 if ((nb+1)%100000==0) fprintf(stderr,"En cours : %d\n",nb+1);
 graf2=strtok(ch," \t\n");
 sprintf(bigramme,"%s %s",graf1,graf2);
 Tree1g=lia_ajoute_element_avl(Tree1g,graf2,1,NULL); 
 Tree2g=lia_ajoute_element_avl(Tree2g,bigramme,1,NULL); 
 strcpy(graf1,graf2);
 }
fclose(file);
}
 

int main(int argc, char **argv)
{
char ch[TailleLigne],*graf1,*graf2,*proba;
int nb,etat=0;
FILE *file;
double pb;

if (argc<2)
 {
 fprintf(stderr,"Syntaxe : %s <fich corpus> <arpa CMU>\n",argv[0]);
 exit(0);
 }

construit_arbre_gram(argv[1]);

if (!(file=fopen(argv[2],"rt"))) ERREUR("can't open :",argv[2]);

for(nb=0;(fgets(ch,TailleLigne,file))&&((etat==0)||(strncmp(ch,"\\2-grams:",9)));nb++)
 {
 printf("%s",ch);
 if (!strncmp(ch,"\\data\\",6)) etat=1;
 }
printf("%s",ch);
while(fgets(ch,TailleLigne,file))
 {
 if (ch[0]=='-')
  {
  proba=strtok(ch," \t\n"); graf1=strtok(NULL," \t\n"); graf2=strtok(NULL," \t\n");
  if (!strncmp(proba,"-99.",4)) printf("%.4lf ",calcule_proba(graf1,graf2));
  else printf("%s ",proba);
  printf("%s %s\n",graf1,graf2);
  }
 else printf("%s",ch);
 }

fclose(file);
}
 
