/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*  Produit un corpus pour le toolkit CMU a partir
    d'un fichier phon decrit par niveaux et d'un descripteur
    Avec l'option -v, produit un fichier vocabulaire  */
/*  FRED 0500 - Modif 0301 (on vire DEBUT et ETIQ)
                Modif 0302 (on vire les symboles XX)  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <lia_tree_avl.h>

#define TailleLigne	40000

char Lex_ML[TailleLigne];

int NbNiveaux,IndiceML;

void ERREUR(ch1,ch2)
 char *ch1,*ch2;
{
fprintf(stderr,"ERREUR : %s %s\n",ch1,ch2);
exit(0);
}

lia_avl_t affiche_item(char *refe,char *ch,lia_avl_t racine,int si_phrase,int si_phrase_id, int without_cc)
{
int i,deja;
static char ch2[TailleLigne];
char *pt;

if (si_phrase_id) printf("%s",refe);
deja=False;
strcpy(ch2,ch);
pt=strtok(ch," \t\n");
while (pt)
 {
 for(i=0;i<NbNiveaux;i++)
  {
  if (i==IndiceML)
   {
   /* modif 0301 (on vire DEBUT et ETIQ) */
   if ((strcmp(pt,"DEBUT"))&&(strcmp(pt,"ETIQ"))&&(strcmp(pt,"XX"))&&
       ((without_cc==False)||((strcmp(pt,"<s>"))&&(strcmp(pt,"</s>")))))
    {
    if ((si_phrase_id)||((si_phrase)&&(deja))) printf(" %s",pt);
    else
     if (si_phrase) printf("%s",pt);
     else printf("%s\n",pt);
    racine=lia_ajoute_element_avl(racine,pt,True,NULL);
    deja=True;
    }
   }
  if ((pt=strtok(NULL," \t\n"))==NULL)
   if (i<(NbNiveaux-1))  ERREUR("Mauvais format ds le fichier corpus :",ch2);
  }
 }
if ((si_phrase)||(si_phrase_id)) printf("\n");
return racine;
}

void LectureDescripteurFormat(char *nomfic, int niveau)
{
FILE *file;
char ch[TailleLigne],*pt;
int i;

NbNiveaux=0;
IndiceML=niveau;
Lex_ML[0]='\0';

if (!(file=fopen(nomfic,"rt"))) ERREUR("can't open : ",nomfic);
while (fgets(ch,TailleLigne,file))
 {
 if (!strncmp(ch,"NB_NIVEAUX",10))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%d",&NbNiveaux)!=1) ERREUR("Mauvais format1 dans le descripteur","");
  } else
 if ((niveau==0)&&(!strncmp(ch,"INDICE_ML",9)))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%d",&IndiceML)!=1) ERREUR("Mauvais format2 dans le descripteur","");
  } else
 if (!strncmp(ch,"LEXIQUE",7))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if ((strcmp(pt,"REF"))&&(strcmp(pt,"ETI")))
   {
   if (sscanf(pt,"%d",&i)!=1) ERREUR("Mauvais format ds le descripteur","");
   if ((IndiceML)&&(IndiceML==i))
    { pt=strtok(NULL," \t\n"); strcpy(Lex_ML,pt); }
   }
  }
 }
if (NbNiveaux==0) ERREUR("Il manque le nb de niveaux dans le descripteur","");
if (IndiceML==0)  ERREUR("Il manque l'indice du ML dans le descripteur","");
if (Lex_ML[0]=='\0') ERREUR("Il manque l'emplacement du fichier lexique dans le descripteur","");
IndiceML--;
}

int main(int argc, char **argv)
{
char ch[TailleLigne],*pt,refe[TailleLigne],*vocab;
int nb,si_phrase,si_phrase_id,i,if_v_ref,if_v_corpus,without_cc,niveau=0;
lia_avl_t racine=NULL;
FILE *file,*file_lex;

if ((argc==1)||(!strcmp(argv[1],"-h")))
 {
 fprintf(stderr,"Syntaxe : %s <nom du fichier descripteur> [-v_corpus fich_vocab]/\
[-v_ref fich_vocab\n]/[-phrase_id]/[-phrase] [-niveau niveau (default=ML)]",
	argv[0]);
 exit(0);
 }

si_phrase=si_phrase_id=if_v_ref=if_v_corpus=without_cc=False;
niveau=0;

for(nb=1;nb<argc;nb++)
 if (!strcmp(argv[nb],"-v_corpus"))
  if (nb+1==argc) ERREUR("there should be a filename after option '-v_corpus'","");
  else { vocab=argv[++nb]; if_v_corpus=True; }
 else
 if (!strcmp(argv[nb],"-v_ref"))
  if (nb+1==argc) ERREUR("there should be a filename after option '-v_ref'","");
  else { vocab=argv[++nb]; if_v_ref=True; }
 else
 if (!strcmp(argv[nb],"-phrase_id")) si_phrase_id=True;
 else
 if (!strcmp(argv[nb],"-phrase")) si_phrase=True;
 else
 if (!strcmp(argv[nb],"-without_cc")) without_cc=True;
 else
 if (!strcmp(argv[nb],"-niveau"))
  {
  if (nb+1==argc) ERREUR("there should be a value after option '-niveau'","");
  else
   if (sscanf(argv[++nb],"%d",&niveau)!=1) ERREUR("bad value after option '-niveau':",argv[nb]);
  }

LectureDescripteurFormat(argv[1],niveau);

for(nb=0;fgets(ch,TailleLigne,stdin);nb++) if ((ch[0]!='@')&&(ch[0]!='&'))
 {
 if ((nb+1)%100000==0) fprintf(stderr,"En cours : %d\n",nb+1);
 for(i=0;(ch[i])&&(ch[i]!=' ')&&(ch[i]!='\t');i++) refe[i]=ch[i];
 if (ch[i]=='\0') ERREUR("Mauvais format dans la ligne :",ch);
 refe[i]='\0';
 for(++i;(ch[i])&&((ch[i]==' ')||(ch[i]=='\t'));i++);
 racine=affiche_item(refe,ch+i,racine,si_phrase,si_phrase_id,without_cc);
 }

if ((if_v_ref)||(if_v_corpus))
 {
 if (!(file=fopen(vocab,"wt"))) ERREUR("can't write in",vocab);
 if (if_v_corpus) lia_affiche_avl_simple(racine,file);
 else
  {
  if (!(file_lex=fopen(Lex_ML,"rt"))) ERREUR("can't open :",Lex_ML);
  while(fgets(ch,TailleLigne,file_lex))
   {
   /* on ecrit que le mot, pas le code */
   pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
   if ((strcmp(pt,"<UNK>"))&&(strcmp(pt,"DEBUT"))&&(strcmp(pt,"ETIQ")))
    if (!strcmp(pt,"##"))
     {
     system("banner WARNING");
     fprintf(stderr,"ATTENTION : le symbole '##' est considere comme un commentaire dans le \n\
     toolkit du CMU, en consequence il est ignoree dans la construction des ML\n");
     }
    else fprintf(file,"%s\n",pt);
   }
  fclose(file_lex);
  }
 fclose(file);
 }

}
  
