/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  Exploite un arbre produit par 'lia_sct_build'
    Les modes de fonctionnement sont les suivants :
     - General
      * affichage de l'arbre sous forme de pattern
     - Mode SUPERVISE
      * calcule un score sur un fichier d'exemples etiquete
      * etiquette un fichier d'exemple non-etiquete
     - Mode NON SUPERVISE
      * ecris les exemples dans des fichiers correspondant a chaque feuille
*/

/*  Fred 0398 - Modif 0699  - Modif 0100  - Modif 0500  -  Modif 1200 (devt)  
                Modif AT&T : 0901 Repartition of the tests according to the class labels  */

#include <lia_sct.h>
#include <lia_sct_use_lib.h>

int main(int argc, char **argv)
{
ty_node racine;
char *choix,*fich_corpus,nom_arbre[TailleLigne];
int min_yes=-1,n,just_correct=False,print_repart=False;
ty_pattern pattern;
double seuil=0;
ty_format my_format;

if ((argc==1)||(!strcmp(argv[1],"-h")))
 {
 fprintf(stderr,"Syntaxe : %s [-h] -desc <fichier descripteur> \
<-corpus_feuille <rep fich corpus .exp>/-test <seuil>/-test_mult <seuil>/-etiquette <seuil>/-affiche> \
[-min_exp nb] [-min_yes nb] [-print_repart] [-just_correct]\n",argv[0]);
 exit(0);
 }

fprintf(stderr,"Chargement des modeles -> ");

for (n=1;n<argc;n++)
 if (!strcmp(argv[n],"-desc"))
  if (n+1<argc) my_format=lis_descripteur(argv[++n],&racine,nom_arbre);
  else ERREUR("descripteur missing","");
 else
 if (!strcmp(argv[n],"-corpus_feuille"))
  if (n+1<argc) { choix=argv[n]; fich_corpus=argv[++n]; }
  else ERREUR("corpus_feuille missing","");
 else
 if (!strcmp(argv[n],"-test"))
  if (n+1<argc) { choix=argv[n]; if (sscanf(argv[++n],"%lf",&seuil)!=1) ERREUR("bad threshold:",argv[n]); }
  else ERREUR("threshold missing","");
 else
 if (!strcmp(argv[n],"-test_mult"))
  if (n+1<argc) { choix=argv[n]; if (sscanf(argv[++n],"%lf",&seuil)!=1) ERREUR("bad threshold:",argv[n]); }
  else ERREUR("threshold missing","");
 else
 if (!strcmp(argv[n],"-etiquette"))
  if (n+1<argc) { choix=argv[n]; if (sscanf(argv[++n],"%lf",&seuil)!=1) ERREUR("bad threshold:",argv[n]); }
  else ERREUR("threshold missing","");
 else
 if (!strcmp(argv[n],"-affiche"))
  { choix=argv[n]; if (argc>n+1) if (sscanf(argv[++n],"%lf",&seuil)!=1) ERREUR("bad threshold:",argv[n]); } else
  /* permet de ne pas stocker d'exemple qui n'ont pas repondu 'oui'
     a au moins min_yes question */
 if (!strcmp(argv[n],"-min_yes"))
  if ((n+1<argc)&&(sscanf(argv[++n],"%d",&min_yes)==1)) ;
  else ERREUR("threshold missing","");
 else
 if (!strcmp(argv[n],"-just_correct")) just_correct=True;
 else
 if (!strcmp(argv[n],"-print_repart")) print_repart=True;
 else
  {
  fprintf(stderr,"ERREUR : je ne connais pas l'option '%s' ....\n",argv[n]);
  exit(0);
  }

fprintf(stderr,"Termine\n");

if ((!strcmp(choix,"-corpus_feuille"))||(!strcmp(choix,"-test"))||(!strcmp(choix,"-test_mult"))||(!strcmp(choix,"-etiquette")))
 {
 traite_fich_exemple(choix,stdin,racine,fich_corpus,min_yes,seuil,my_format,just_correct,print_repart);
 }
else
 {
 if (!strcmp(choix,"-affiche"))
  {
  pattern=nouveau_pattern("+",NULL,NULL,NULL);
  pattern->pred=pattern->suiv=NULL;
  affiche_pattern(racine,pattern,pattern,seuil,my_format);
  }
 }
exit(0);
}
  
