/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  Exploite un arbre produit par 'lia_sct_build'
    Les modes de fonctionnement sont les suivants :
     - General
      * affichage de l'arbre sous forme de pattern
     - Mode SUPERVISE
      * calcule un score sur un fichier d'exemples etiquete
      * etiquette un fichier d'exemple non-etiquete
     - Mode NON SUPERVISE
      * ecris les exemples dans des fichiers correspondant a chaque feuille
*/

/*  Fred 0398 - Modif 0699  - Modif 0100  - Modif 0500  -  Modif 1200 (devt)  
                Modif AT&T : 0901 Repartition of the tests according to the class labels
		Modif AT&T : 1001 Making a library of 'lia_sct_use' into 'lia_sct_use_lib'
				  Adding a function : traverse_string which takes a simple
				  'char *' of words and traverse the tree with it and return
				  the probability distribution of the leaf reached
*/

#include <lia_sct.h>
#include <lia_sct_use_lib.h>

/*................................................................*/

/*  Codage de la repartition des exemples apres le parcours de l'arbre  */

typedef struct
	{
	ty_node feuille;
	int nb;
	FILE *file;
	} type_groupe;

int NbGroupe;
type_groupe *TableGroupe;

/*................................................................*/

int top_categorie(double *proba,double *maxproba,ty_format my_format)
{
int i,indice;
double max;
for(i=1,indice=0,max=proba[0];i<my_format->nb_etiquette;i++)
 if (proba[i]>max) { max=proba[i]; indice=i; }
*maxproba=max;
return indice;
}

/*................................................................*/

/* affichage des patterns : c'est a dire des expressions
   reguliere (incomplete, car sans les negations) */

ty_pattern nouveau_pattern(char *graf, char *graf_glb, ty_pattern pred, ty_pattern suiv)
{
ty_pattern pt;
pt=(ty_pattern)malloc(sizeof(struct type_pattern));
if (graf) pt->graf=(char*)strdup(graf); else pt->graf=NULL;
if (graf_glb) pt->graf_glb=(char*)strdup(graf_glb); else pt->graf_glb=NULL;
pt->pred=pred;
pt->suiv=suiv;
return pt;
}

void libere_pattern(pt)
 ty_pattern pt;
{
if (pt->graf) free(pt->graf);
free(pt);
}

void libere_list_pattern(pt)
 ty_pattern pt;
{
ty_pattern ptbak;

while(pt)
 {
 ptbak=pt;
 pt=pt->suiv;
 libere_pattern(ptbak);
 }
}

/*................................................................*/

/*  Fonctions d'affichage  */

int si_trou_droite(ty_pattern pt) /* test la presence d'un trou + a droite de pt */
{
return pt->graf[strlen(pt->graf)-1]=='+'?1:0;
}

int si_trou_gauche(ty_pattern pt) /* test la presence d'un trou + a gauche de pt */
{
return pt->graf[0]=='+'?1:0;
}

int si_trou(ty_pattern pt) /* test la presence d'un trou + + entre pt et pt->suiv */
{
return ((si_trou_droite(pt))&&((pt->suiv==NULL)||(si_trou_gauche(pt->suiv))))?1:0;
}

void affiche_pattern(ty_node pt,ty_pattern racphr,ty_pattern ptph,double seuil,ty_format my_format)
{
static char ch[200],graf[200];
static char ch_glb[TailleLigne];
ty_pattern tmp_av,tmp_pt,tmp_ap;
static int i=-1;
char *ch_q;
int indice;
double maxproba;

if (ptph==NULL) ERREUR("WHY ptph==NULL","");

if (pt)
 {
 if (pt->t_noeud=='L')
  {
  i++;
  if (pt->proba!=NULL) indice=top_categorie(pt->proba,&maxproba,my_format);

  if ((pt->proba==NULL)||(maxproba>=seuil))
   {
   printf("\nCorpus %d - NODE_ID=%d - Pattern : ",i,pt->cate);
   for(ch_glb[0]='\0',tmp_pt=racphr->suiv;tmp_pt;tmp_pt=tmp_pt->suiv)
    {
    if (tmp_pt->graf)
     {
     if ((si_trou_gauche(tmp_pt))&&(tmp_pt->pred!=racphr)) strcpy(graf,tmp_pt->graf+2); else strcpy(graf,tmp_pt->graf);
     if ((si_trou_droite(tmp_pt))&&(tmp_pt->suiv)&&(!si_trou_gauche(tmp_pt->suiv))) graf[strlen(graf)-2]='\0';
     printf("%s ",graf);
     }
    if (tmp_pt->graf_glb) sprintf(ch_glb+strlen(ch_glb)," %s",tmp_pt->graf_glb);
    }
   if (ch_glb[0]) printf(" |%s",ch_glb);
   if (pt->proba) printf(" => %s (%2.4lf)",renvoi_chaine_lex(indice+1,my_format->lex_etiquette),maxproba);
   printf("\n");
   }
  }
 else
  if (pt->t_noeud=='I')
   {
   ch_q=renvoi_chaine(pt->q,pt->t_question=='G'?my_format->nb_niveaux:pt->niveau_q,my_format);
   switch (pt->t_question)
    {
    case 'J' : sprintf(ch,"%s",ch_q); break;
    case 'L' : sprintf(ch,"%s +",ch_q); break;
    case 'R' : sprintf(ch,"+ %s",ch_q); break;
    case 'U' : sprintf(ch,"+ %s +",ch_q); break;
    case 'G' : sprintf(ch,"<%s>",ch_q); break;
    }
   if (ptph==racphr)
    {
    tmp_av=racphr; tmp_ap=NULL;
    if (pt->t_question=='G') tmp_pt=racphr->suiv=nouveau_pattern(NULL,ch,racphr,NULL);
    else tmp_pt=racphr->suiv=nouveau_pattern(ch,NULL,racphr,NULL);
    }
   else
    {
    if (pt->t_question!='G')
     {
     if (ptph->graf==NULL)
      {
      ptph->graf=strdup(ch);
      tmp_av=ptph->pred;
      tmp_ap=ptph->suiv;
      tmp_pt=ptph;
      }
     else
      {
      if (pt->qloc=='l')
       for(tmp_av=ptph->pred;(tmp_av)&&(!si_trou(tmp_av));tmp_av=tmp_av->pred);
      else
       for(tmp_av=ptph;(tmp_av)&&(!si_trou(tmp_av));tmp_av=tmp_av->suiv);
      tmp_ap=tmp_av->suiv;
      tmp_pt=nouveau_pattern(ch,NULL,tmp_av,tmp_ap);
      tmp_av->suiv=tmp_pt;
      if (tmp_ap) tmp_ap->pred=tmp_pt;
      }
     }
    else
     {
     if (ptph->graf_glb==NULL) ptph->graf_glb=strdup(ch);
     else
      {
      sprintf(ch_glb,"%s %s",ptph->graf_glb,ch);
      free(ptph->graf_glb);
      ptph->graf_glb=strdup(ch_glb);
      }
     tmp_av=ptph->pred;
     tmp_ap=ptph->suiv;
     tmp_pt=ptph;
     }
    if (tmp_av==NULL) ERREUR("Incoherence dans l'arbre","");
    }
   affiche_pattern(pt->fg,racphr,tmp_pt,seuil,my_format);
   if (tmp_pt!=ptph)
    {
    tmp_av->suiv=tmp_ap;
    if (tmp_ap) tmp_ap->pred=tmp_av;
    libere_pattern(tmp_pt);
    }
   else
    {
    if ((pt->t_question=='G')&&(tmp_pt->graf_glb))
     {
     int k;
     for(k=strlen(tmp_pt->graf_glb)-1;(k>0)&&(tmp_pt->graf_glb[k]!=' ');k--);
     if (k>0) tmp_pt->graf_glb[k]='\0';
     else { free(tmp_pt->graf_glb); tmp_pt->graf_glb=NULL; }
     }
    else
     {
     if (tmp_pt->graf) free(tmp_pt->graf);
     tmp_pt->graf=NULL;
     }
    }
   affiche_pattern(pt->fd,racphr,ptph,seuil,my_format);
   }
 }
}

/*................................................................*/

/*  Parcours du CART avec un item  */

#define MaxItem		10000

void SauvEtat(pt,etat)
 ty_item pt;
 char *etat;
{
int n;
for(n=0;(n<MaxItem)&&(pt);pt=pt->suiv,n++)
 etat[n]=pt->inKS;
if (n==MaxItem) ERREUR("Cste MaxItem trop petite !!!!","");
}

void RestorEtat(pt,etat)
 ty_item pt;
 char *etat;
{
int n;
for(n=0;(n<MaxItem)&&(pt);pt=pt->suiv,n++)
 pt->inKS=etat[n];
if (n==MaxItem) ERREUR("Cste MaxItem trop petite !!!!","");
}

ty_item Compatible(item,KS,question,niveau_q,qloc,gap_avant,gap_apres)
 ty_item item;
 type_KS KS;
 wrd_index_t question;
 char niveau_q,qloc,gap_avant,gap_apres;
{
ty_item pt,tmp;

if (qloc=='l')  /* on cherche a la gauche du KS */
 {   /* sur la racine, ptKS est egal a NULL */

 /* test de coherence */
 /* s'il n'y a pas de GAP avant, on recule vers la gauche jusqu'a en trouver un */
 if (KS.gap_avant==0)
  {
  for(pt=KS.ptKS;(pt)&&(pt->inKS==1);pt=pt->pred);
  if (pt==NULL) return NULL; /*ERREUR("Il n'y a pas de GAP sur la gauche !!!!","");*/
  }
 else if (KS.ptKS==NULL) pt=item; else pt=KS.ptKS->pred;

 if (pt==NULL) return NULL;

 if (gap_apres==0)
  { /* la question doit se trouver juste avant le KS */
  if ((pt->inKS!=0)||(question!=pt->code[niveau_q])) return NULL;
  }
 else
  { /* il peut y avoir un gap entre la question et le KS */
  /* MODIF : il DOIT y avoir un gap */
  pt=pt->pred;
  while ((pt)&&(pt->inKS==0)&&(question!=pt->code[niveau_q])) pt=pt->pred;
  if ((pt==NULL)||(pt->inKS!=0)) return NULL;
  }
 /* la solution est dans 'pt', verifions le gap_avant */
 if (gap_avant==0) /* il ne faut pas qu'il y est de trou avant */ 
  {
  for(tmp=pt->pred;(tmp)&&(tmp->inKS==1);tmp=tmp->pred);
  if (tmp!=NULL) return NULL; /* il ne fallait pas un trou avant */
  }
 pt->inKS=1;
 return pt;
 }
else
 {  /* on cherche a la droite du KS */

 /* test de coherence */
 /* s'il n'y a pas de GAP apres, on avance vers la droite jusqu'a en trouver un */
 if (KS.gap_apres==0)
  {
  for(pt=KS.ptKS==NULL?item:KS.ptKS;(pt)&&(pt->inKS==1);pt=pt->suiv);
  if (pt==NULL) return NULL; /*ERREUR("Il n'y a pas de GAP sur la droite !!!!","");*/
  }
 else if (KS.ptKS==NULL) pt=item; else pt=KS.ptKS->suiv;

 if (pt==NULL) return NULL;

 if (gap_avant==0)
  { /* la question doit se trouver juste apres le KS */
  if ((pt->inKS!=0)||(question!=pt->code[niveau_q])) return NULL;
  }
 else
  { /* il peut y avoir un gap entre la question et le KS */
  /* MODIF : il DOIT y avoir un gap */
  pt=pt->suiv;
  while ((pt)&&(pt->inKS==0)&&(question!=pt->code[niveau_q])) pt=pt->suiv;
  if ((pt==NULL)||(pt->inKS!=0)) return NULL;
  }
 /* la solution est dans 'pt', verifions le gap_apres */
 if (gap_apres==0) /* il ne faut pas qu'il y est de trou apres */
  {
  for(tmp=pt->suiv;(tmp)&&(tmp->inKS==1);tmp=tmp->suiv);
  if (tmp!=NULL) return NULL; /* il ne fallait pas un trou apres */
  }
 pt->inKS=1;
 return pt;
 }
}

ty_node Parcours(
 ty_node pt,
 ty_item item,
 wrd_index_t *questall,
 type_KS KS,
 int *nb_yes,
 ty_format my_format)
{
char etat[MaxItem];
ty_item resu;
type_KS newKS;
static char Debug_quest[TailleLigne];

if (pt)
 {
 if (pt->t_noeud=='L')
  {
  if (Debug>=2) fprintf(stdout,"Num feuille : %d\n",pt->cate);
  return pt;
  }
 else
  {
  /*  S'il n'y a pas de GAP, on s'arrete  */
  /*
  if ((KS.gap_avant==0)&&(KS.gap_apres==0))
   {
   fprintf(stdout,"Warning : pas de GAP -> queskon renvoit ????\n");
   return;
   }
  */

  SauvEtat(item,etat);

  if (Debug>=2) { fprintf(stdout,"item = "); affiche_item(item,my_format); }
  switch (pt->t_question)
   {
   case 'J' :  /* < ? > */
	if (Debug>=2) sprintf(Debug_quest,"<%s>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,0,0))
	 { newKS.gap_avant=newKS.gap_apres=0; newKS.ptKS=resu; }
	break;
   case 'L' : /* < ?+> */
	if (Debug>=2) sprintf(Debug_quest,"<%s+>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,0,1))
	 { newKS.gap_avant=0; newKS.gap_apres=1; newKS.ptKS=resu; }
	break;
   case 'R' : /* <+? > */
	if (Debug>=2) sprintf(Debug_quest,"<+%s>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,1,0))
	 { newKS.gap_avant=1; newKS.gap_apres=0; newKS.ptKS=resu; }
	break;
   case 'U' : /* <+?+> */
	if (Debug>=2) sprintf(Debug_quest,"<+%s+>",renvoi_chaine(pt->q,pt->niveau_q,my_format));
	if (resu=Compatible(item,KS,pt->q,pt->niveau_q,pt->qloc,1,1))
	 { newKS.gap_avant=newKS.gap_apres=1; newKS.ptKS=resu; }
	break;
   case 'G' : /* global questions */
	if (Debug>=2) sprintf(Debug_quest,"|%s|",renvoi_chaine(pt->q,my_format->nb_niveaux,my_format));
	resu=NULL;
	if (questall)
 	 {
	 int i;
	 for(i=0;(i<my_format->nb_questall)&&(questall[i]!=pt->q);i++);
	 if (i<my_format->nb_questall)
	  {
	  resu=(ty_item)True;
	  newKS.gap_avant=KS.gap_avant; newKS.gap_apres=KS.gap_apres; newKS.ptKS=KS.ptKS;
	  }
	 }
	break;
   default  :
	fprintf(stdout,"Je ne connais pas la question : %c\n",pt->t_question);
	exit(0);
   }
  if (resu)
   {
   if (Debug>=2)
    {
    fprintf(stdout,"  Oui a la question : %s sur le GAP %c\n",Debug_quest,pt->qloc);
    fprintf(stdout,"    avec KS=<%c,%s,%c>\n",'0'+KS.gap_avant,
	KS.ptKS!=NULL?renvoi_chaine((KS.ptKS)->code[my_format->indice_ML],my_format->indice_ML,my_format):"NULL",'0'+KS.gap_apres);
    }
   (*nb_yes)++;
   return Parcours(pt->fg,item,questall,newKS,nb_yes,my_format);
   }
  else
   {
   if (Debug>=2)
    {
    fprintf(stdout,"  Non a la question : %s sur le GAP %c\n",Debug_quest,pt->qloc);
    fprintf(stdout,"    avec KS=<%c,%s,%c>\n",'0'+KS.gap_avant,
	KS.ptKS!=NULL?renvoi_chaine((KS.ptKS)->code[my_format->indice_ML],my_format->indice_ML,my_format):"NULL",'0'+KS.gap_apres);
    fprintf(stdout,"    questall=");
    if (questall)
     {
     int i;
     for(i=0;(i<my_format->nb_questall)&&(questall[i]!=0);i++)
      printf(" %s",renvoi_chaine(questall[i],my_format->nb_niveaux,my_format));
     printf("\n");
     }
    else printf(" NULL\n");
    }
   RestorEtat(item,etat);
   return Parcours(pt->fd,item,questall,KS,nb_yes,my_format);
   }
  }
 }
return NULL;
}

/*................................................................*/

/*  Lecture d'un fichier CART et construction de l'arbre  */

ty_node ConstruitArbre(FILE *file,char KS,ty_format my_format)
{
char ch[TailleLigne],t_noeud,t_question,qloc,q[TailleLigne],old_KS,niveau_q,*etiq,*pb;
ty_node pt;
wrd_index_t code_q,i_etiq;
int i,int_niveau_q,cate;
double *proba;

if (fgets(ch,TailleLigne,file))
 {
 i=sscanf(ch,"%c %d %c %s %d %c %c",&t_noeud,&cate,&old_KS,q,&int_niveau_q,&t_question,&qloc);
 if (i!=7)
  {
  sprintf(q,"%d elements trouves dans : %s (au lieu de 7)",i,ch);
  ERREUR("Mauvais format dans le fichier arbre",q);
  }

 niveau_q=(char)int_niveau_q-1;

 if (!strcmp(q,"NULL")) code_q=0; /* on mets le <UNK> sur NULL */
 else
  if (!Mot2Code(q,&code_q,my_format->tablex[t_question=='G'?my_format->nb_niveaux:niveau_q])) ERREUR("Item inconnu : ",q);

 proba=NULL;

 if (t_noeud=='L') /* on est sur une feuille */
  {
  NbGroupe++;
  if (my_format->mode==SUPERVISE) /* il doit y avoir une repartition de probas */
   {
   if (!fgets(ch,TailleLigne,file)) ERREUR("Pas de repart de proba ds l'arbre","");
   proba=(double*)malloc(sizeof(double)*my_format->nb_etiquette);
   for(i=0;i<my_format->nb_etiquette;i++)
    {
    if (i==0) etiq=strtok(ch," \t\n"); else etiq=strtok(NULL," \t\n");
    if (etiq==NULL) ERREUR("repart de proba non valide","");
    pb=strtok(NULL," \t\n");
    if (pb==NULL) ERREUR("repart de proba non valide","");
    if (!Mot2Code(etiq,&i_etiq,my_format->lex_etiquette)) ERREUR(etiq,"inconnu");
    if (sscanf(pb,"%lf",&(proba[i_etiq-1]))!=1) ERREUR("repart de proba non valide","");
    }
   }
  }

 pt=nouveau_node(KS,t_noeud,cate,t_question,qloc,code_q,niveau_q,proba);

 if (t_noeud=='I')
  {
  pt->fg=ConstruitArbre(file,t_question,my_format);
  pt->fd=ConstruitArbre(file,KS,my_format);
  }
 return pt;
 }
return NULL;
}

void AffecteFeuille(racine,i)
 ty_node racine;
 int *i;
{
if (racine)
 {
 if (racine->t_noeud=='L') /* c'est une feuille */
  {
  if (*i==NbGroupe) { fprintf(stderr,"NbGroupe trop petit !!\n"); exit(0); }
  TableGroupe[*i].feuille=racine;
  TableGroupe[*i].nb=0;
  TableGroupe[*i].file=NULL;
  (*i)++;
  }
 AffecteFeuille(racine->fg,i);
 AffecteFeuille(racine->fd,i);
 }
}

ty_node LectureTree(char *chfile,ty_format my_format)
{
FILE *file;
int n;
ty_node racine;

if (!(file=fopen(chfile,"rt"))) ERREUR("Can't open : ",chfile);

/*  La KS a la racine est X cad <+>  */
NbGroupe=0;
racine=ConstruitArbre(file,'X',my_format);
fclose(file);

/* on alloue le tableau des groupes d'item */
TableGroupe=(type_groupe*)malloc(sizeof(type_groupe)*NbGroupe);
n=0;
AffecteFeuille(racine,&n);
return racine;
}

/*................................................................*/

/*  Gere les groupes d'exemples dans les feuilles  */

void print_repart_labels(char *ref,double *proba,int numnode,FILE *file,ty_format my_format)
{
int i;
fprintf(file,"%s",ref);
for(i=0;i<my_format->nb_etiquette;i++)
 fprintf(file," %s %.2lf",renvoi_chaine_lex(i+1,my_format->lex_etiquette),proba[i]);
fprintf(file,"\t(NODE=%d)\n",numnode);
}

void ecris_item_niveaux(
	char *ref,
	char *etiquette,
	ty_item item,
	FILE *file,
	ty_format my_format)
{
int niveau;
fprintf(file,"%s",ref);
while (item)
 {
 for(niveau=0;niveau<my_format->nb_niveaux;niveau++) 
  fprintf(file," %s_%c",renvoi_chaine(item->code[niveau],niveau,my_format),(char)(item->inKS+'0'));
 item=item->suiv;
 }
fprintf(file," = %s\n",etiquette);
}

/* if the threshold is not reach, then the answer is 'OTHER' with everything at '0' HMIHY1001 */
void ecris_item_niveaux_vide(
	char *ref,
	char *etiquette,
	ty_item item,
	FILE *file,
	ty_format my_format)
{
int niveau;
fprintf(file,"%s",ref);
while (item)
 {
 for(niveau=0;niveau<my_format->nb_niveaux;niveau++)
  fprintf(file," %s_0",renvoi_chaine(item->code[niveau],niveau,my_format));
 item=item->suiv;
 }
fprintf(file," = OTHER\n");
}

void ecris_item(char *ref,ty_item item,FILE *file,ty_format my_format, int if_jump)
{
int i;
fprintf(file,"%s",ref);
while (item)
 {
 for(i=0;i<my_format->nb_niveaux;i++) fprintf(file," %s",renvoi_chaine(item->code[i],i,my_format));
 item=item->suiv;
 }
if (if_jump) fprintf(file,"\n");
}

void ecris_item_etiquette(char *ref,ty_item item,char *etiquette,FILE *file,ty_format my_format)
{
int i;
fprintf(file,"%s",ref);
while (item)
 {
 for(i=0;i<my_format->nb_niveaux;i++) fprintf(file," %s",renvoi_chaine(item->code[i],i,my_format));
 item=item->suiv;
 }
fprintf(file," = %s\n",etiquette);
}

void RajouteExempleDansGroupe(
 char *ref,
 ty_item item,
 wrd_index_t *questall,
 ty_node feuille,
 char *nom_corpus,
 ty_format my_format)
{
int n;
/*static char ch[TailleLigne];*/

for(n=0;(n<NbGroupe)&&(TableGroupe[n].feuille!=NULL)&&(TableGroupe[n].feuille!=feuille);n++);
if (n==NbGroupe) ERREUR("Il y a une feuille de plus que NbGroupe !!","");
TableGroupe[n].feuille=feuille;
TableGroupe[n].nb++;

/*
MODIF HMIHY 12/06/01 -> now we print the IDs and the leaves in the same output file
printf("f%d\t%s\t%s\n",n,ref,renvoi_chaine(item->code[1],1,my_format)); 1 because the first word is the prompt ID
*/
/*
MODIF HMIYH 03/21/02 -> now we print the samples with their filenames at the beginning of each line
                        but we don't write it in order to prevent too many in/out writting: we do it after
*/
printf("%s/feuille%d.exp\t",nom_corpus,n);
ecris_item(ref,item,stdout,my_format,False);
if (questall)
 {
 int i;
 printf(" &&&&");
 for(i=0;(i<my_format->nb_questall)&&(questall[i]!=0);i++)
 printf(" %s",renvoi_chaine(questall[i],my_format->nb_niveaux,my_format));
 printf("\n");
 }

/*
sprintf(ch,"%s/feuille%d.exp",nom_corpus,n);
TableGroupe[n].file=fopen(ch,"a");
if (!TableGroupe[n].file) { fprintf(stderr,"can't open : %s\n",ch); exit(0); }
ecris_item(ref,item,TableGroupe[n].file,my_format,True);
fclose(TableGroupe[n].file);
*/
libere_list_item(item);
}

/*................................................................*/

/* Lecture du corpus et construction des structures des ngrams */
/* MODIF : lecture de fichiers au format :

i=item (l items), n=niveau (k niveaux)

<graf> <i1 n1> <i1 n2> .... <i1 nk> <i2 n1> <i2 n2> .... <i2 nk> .... <ilnk>

*/

ty_node traite_item(char *choix, char *ref, ty_item item, wrd_index_t *questall,
		    ty_node racine, int min_yes, char *nom_corpus,ty_format my_format)
{
ty_node feuille;
type_KS KS;
int nb_yes;

if (Debug>=2)
 { printf("On traite : %s ",ref); if (item) affiche_item(item,my_format); else printf("NULL ????\n"); }

/*  Initialisation du KS  */
KS.gap_avant=KS.gap_apres=1;
KS.ptKS=NULL;

/*  Parcours de l'arbre de decision  */
nb_yes=0;
feuille=Parcours(racine,item,questall,KS,&nb_yes,my_format);

if (feuille==NULL)
 fprintf(stderr,"L'item suivant n'est pas analysee par le CART : %s\n",ref);
else
 if (!strcmp(choix,"-corpus_feuille"))
  if (nb_yes>=min_yes) RajouteExempleDansGroupe(ref,item,questall,feuille,nom_corpus,my_format);
  else
   {
   if (Debug>=2) printf("Exemple : %s\n\t=> non rajoute car : nb_yes=%d et min_yes=%d\n",ref,nb_yes,min_yes);
   libere_list_item(item);
   }
return feuille;
}

/*................................................................*/

/*  MODIF ATT - october 2001 : 'traverse_string'
-> function which takes a sentence as an 'ty_item' chained list and traverse the
tree with it and return the probability distribution of the leaf reached  */

double *traverse_string(ty_item item, ty_node racine,ty_format my_format)
{
type_KS KS;
int nb_yes;
ty_node feuille;

/*  Initialisation du KS  */
KS.gap_avant=KS.gap_apres=1;
KS.ptKS=NULL;
/*  Parcours de l'arbre de decision  */
nb_yes=0;
feuille=Parcours(racine,item,NULL,KS,&nb_yes,my_format);
if (feuille==NULL) ERREUR("the item is not analized by the SCT","");
return feuille->proba;
}

/*................................................................*/

/* lis un fichier d'exemple en le faisant passer dans l'arbre */

void traite_fich_exemple(
	char *choix,
	FILE *file,
	ty_node racine,
	char *fich_corpus,
	int min_yes,
	double seuil,
	ty_format my_format,
	int just_correct,
	int print_repart
	)
{
int n,i,top,nbgood=0,nbbaad=0,*repart_good,*repart_baad,*repart_refe,nb_test=0;
char ch[10*TailleLigne],*graf,refe_id[TailleLigne],*chitem,*etiq,*pafini;
ty_node feuille;
ty_item item;
wrd_index_t i_etiq;
double maxproba;
wrd_index_t *questall;
float prec,recall,fmes;

if (my_format->nb_questall)
 questall=(wrd_index_t *)malloc(sizeof(wrd_index_t)*my_format->nb_questall);
else
 questall=NULL;

if ((!strcmp(choix,"-test"))||(!strcmp(choix,"-test_mult")))
 {
 repart_good=(int*)malloc(sizeof(int)*my_format->nb_etiquette);
 repart_baad=(int*)malloc(sizeof(int)*my_format->nb_etiquette);
 repart_refe=(int*)malloc(sizeof(int)*my_format->nb_etiquette);
 for(i=0;i<my_format->nb_etiquette;i++) repart_good[i]=repart_baad[i]=repart_refe[i]=0;
 }

for(n=0,pafini=fgets(ch,10*TailleLigne,file);pafini;n++)
 if (ch[0]!='@')
  {
  if (ch[0]=='&') ERREUR("Wrong format in string:",ch);
  if ((n+1)%5000==0) fprintf(stderr,"En cours : %d\n",n+1);
  graf=(char*)strtok(ch," \t\n");
  for(i=strlen(graf)+1;(graf[i]==' ')||(graf[i]=='\t');i++); chitem=graf+i;
  strcpy(refe_id,graf);
  item=CodeItem(chitem,my_format);
  pafini=fgets(ch,10*TailleLigne,file);
  if ((pafini)&&(ch[0]=='&'))
   {
   for(i=0,graf=strtok(ch+2," \t\n");graf;graf=strtok(NULL," \t\n"))
    {
    if (i>=my_format->nb_questall) ERREUR("too much global questions in sample:",refe_id);
    if (Mot2Code(graf,&(questall[i]),my_format->tablex[my_format->nb_niveaux])) i++;
    else fprintf(stderr,"Warning: glb question not in the glb lexicon:%s\n",graf);
    }
   for(;i<my_format->nb_questall;i++) questall[i]=0;
   pafini=fgets(ch,10*TailleLigne,file);
   }
  if (feuille=traite_item(choix,refe_id,item,questall,racine,min_yes,fich_corpus,my_format))
   {
   if (strcmp(choix,"-corpus_feuille"))
    {
    if (feuille->proba==NULL) ERREUR("la feuille n'as pas de proba !!",refe_id);
    top=top_categorie(feuille->proba,&maxproba,my_format);
    nb_test++;
    if (!strcmp(choix,"-etiquette"))
     {
     if (feuille->proba[top]>=seuil)
      {
      if (print_repart)
       {
       /*ecris_item_etiquette(refe_id,item,renvoi_chaine_lex(top+1,my_format->lex_etiquette),stdout,my_format);*/
       print_repart_labels(refe_id,feuille->proba,feuille->cate,stdout,my_format);
       }
      /*
      fprintf(stdout,"@ %s 1\n",renvoi_chaine_lex(top+1,my_format->lex_etiquette));
      */

      if (just_correct)
       {
       if ((!pafini)||(ch[0]!='@')) ERREUR("Pas de categorie en mode 'just correct'","");
       etiq=strtok(ch," \t\n"); etiq=strtok(NULL," \t\n");
       if (!Mot2Code(etiq,&i_etiq,my_format->lex_etiquette)) ERREUR(etiq,"unknown");
       pafini=fgets(ch,10*TailleLigne,file);
       }
      if ((print_repart==False)&&((just_correct==False)||(i_etiq==top+1)))
       {
       /*ecris_item_niveaux(refe_id,renvoi_chaine_lex(top+1,my_format->lex_etiquette),item,stdout,my_format);*/

       ecris_item_etiquette(refe_id,item,renvoi_chaine_lex(top+1,my_format->lex_etiquette),stdout,my_format);
       }
      }
     else /* MODIF HMIHY 1001 */
      {
      if (print_repart)
       {
       ecris_item_etiquette(refe_id,item,"OTHER",stdout,my_format);
       print_repart_labels(refe_id,feuille->proba,feuille->cate,stdout,my_format);
       }
      else ecris_item_niveaux_vide(refe_id,renvoi_chaine_lex(top+1,my_format->lex_etiquette),item,stdout,my_format);
      }
     }
    else
     if (!strcmp(choix,"-test"))
      {
      if ((!pafini)||(ch[0]!='@')) ERREUR("Pas de categorie en mode test","");
      etiq=strtok(ch," \t\n"); etiq=strtok(NULL," \t\n");

      if (!Mot2Code(etiq,&i_etiq,my_format->lex_etiquette)) ERREUR(etiq,"unknown");
      repart_refe[i_etiq-1]++;
      
      if (feuille->proba[top]>=seuil)
       {
       if (i_etiq==top+1)
        { nbgood++; repart_good[top]++; }
       else
        { nbbaad++; repart_baad[top]++; }
       }
      pafini=fgets(ch,10*TailleLigne,file);
      }
	 else
     if (!strcmp(choix,"-test_mult"))
      {
      if ((!pafini)||(ch[0]!='@')) ERREUR("Pas de categorie en mode test","");
	  etiq=strtok(ch," \t\n");
      for (etiq=strtok(NULL," \t\n");etiq;etiq=strtok(NULL," \t\n"))
       if (Mot2Code(etiq,&i_etiq,my_format->lex_etiquette))
        {
	    repart_refe[i_etiq-1]++;
	    if (feuille->proba[i_etiq-1]>=seuil)
         { nbgood++; repart_good[i_etiq-1]++; }
	    else
        { nbbaad++; repart_baad[i_etiq-1]++; }
        }
	  pafini=fgets(ch,10*TailleLigne,file);
      }
    }
   }
  }
 else pafini=fgets(ch,10*TailleLigne,file);

if ((!strcmp(choix,"-test"))||(!strcmp(choix,"-test_mult")))
 {
 prec=nbgood+nbbaad>0?(float)(nbgood*100)/(float)(nbgood+nbbaad):0;
 recall=nb_test>0?(float)(nbgood*100)/(float)nb_test:100;
 fmes=(prec+recall)>0?(prec*recall*(float)2)/(prec+recall):0;
 fprintf(stdout,"Results: nb of test=%d - nb of answer=%d - correct=%d - precision=%2.2f%% - recall=%2.2f%% - F=%2.2f\n",
  nb_test,nbgood+nbbaad,nbgood,prec,recall,fmes);
 fprintf(stdout,"\tdetails:\n");
 for(i=0;i<my_format->nb_etiquette;i++)
  {
  prec=(repart_good[i]+repart_baad[i])>0?(float)(100*repart_good[i])/(float)(repart_good[i]+repart_baad[i]):0;
  recall=repart_refe[i]>0?(float)(100*repart_good[i])/(float)repart_refe[i]:100;
  fmes=(prec+recall)>0?(prec*recall*(float)2)/(prec+recall):0;
  printf("\t -> %16s - #decision=%4d - #ok=%4d - #ref=%4d - prec=%2.2f%% - recall=%2.2f%% - F=%2.2f%%\n",
   renvoi_chaine_lex(i+1,my_format->lex_etiquette),
   repart_good[i]+repart_baad[i],repart_good[i],repart_refe[i],prec,recall,fmes);
  }
 free(repart_good); free(repart_baad); free(repart_refe);
 }

if (questall) free(questall);
}

/*................................................................*/

ty_format lis_descripteur(char *fich_desc,ty_node *racine,char *nom_arbre)
{
FILE *file;
char ch[TailleLigne],*pt,ch2[TailleLigne];
int si_arbre=0;
ty_format my_format;

if (!(file=fopen(fich_desc,"rt"))) { fprintf(stderr,"can't open : %s\n",fich_desc); exit(0); }
while(fgets(ch,TailleLigne,file))
 {
 if (!strncmp(ch,"ARBRE",5))
  {
  si_arbre=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  sprintf(ch2,"%s.desc_lex",pt);
  my_format=LectureDescripteurFormat(ch2);
  (*racine)=LectureTree(pt,my_format);
  if (nom_arbre!=NULL) strcpy(nom_arbre,pt);
  }
 }
if (!si_arbre) { fprintf(stderr,"ERREUR : il manque 'ARBRE' dans %s\n",fich_desc); exit(0); }
fclose(file);

return my_format;
}
  
